# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for corpus synchronization with GCS."""

import os
import re
import shutil

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

try:
  from clusterfuzz._internal.google_cloud_utils import gsutil

  # Disable "invalid-name" because fixing the issue will cause pylint to
  # complain the None assignment is incorrectly named.
  DEFAULT_GSUTIL_RUNNER = gsutil.GSUtilRunner  # pylint: disable=invalid-name
except:
  # This is expected to fail on App Engine.
  gsutil = None
  DEFAULT_GSUTIL_RUNNER = None

BACKUP_ARCHIVE_FORMAT = 'zip'
CORPUS_FILES_SYNC_TIMEOUT = 60 * 60
LATEST_BACKUP_TIMESTAMP = 'latest'
PUBLIC_BACKUP_TIMESTAMP = 'public'
REGRESSIONS_GCS_PATH_SUFFIX = '_regressions'

RSYNC_ERROR_REGEX = (br'CommandException:\s*(\d+)\s*files?/objects? '
                     br'could not be copied/removed')

MAX_SYNC_ERRORS = 10


def _rsync_errors_below_threshold(gsutil_result, max_errors):
  """Check if the number of errors during rsync is lower than our threshold."""
  match = re.search(RSYNC_ERROR_REGEX, gsutil_result.output, re.MULTILINE)
  if not match:
    return False

  error_count = int(match.group(1))

  # Ignore NotFoundException(s) since they can happen when files can get deleted
  # e.g. when pruning task is updating corpus.
  error_count -= gsutil_result.output.count(b'NotFoundException')
  error_count -= gsutil_result.output.count(b'No such file or directory')

  return error_count <= max_errors


def _handle_rsync_result(gsutil_result, max_errors):
  """Handle rsync result."""
  if gsutil_result.return_code == 0:
    sync_succeeded = True
  else:
    logs.log_warn(
        'gsutil rsync got non-zero:\n'
        'Command: %s\n'
        'Output: %s\n' % (gsutil_result.command, gsutil_result.output))
    sync_succeeded = _rsync_errors_below_threshold(gsutil_result, max_errors)

  return sync_succeeded and not gsutil_result.timed_out


def _count_corpus_files(directory):
  """Count the number of corpus files."""
  return shell.get_directory_file_count(directory)


def backup_corpus(backup_bucket_name, corpus, directory):
  """Archive and store corpus as a backup.

  Args:
    backup_bucket_name: Backup bucket.
    corpus: The FuzzTargetCorpus.
    directory: Path to directory to be archived and backuped.

  Returns:
    The backup GCS url, or None on failure.
  """
  if not backup_bucket_name:
    logs.log('No backup bucket provided, skipping corpus backup.')
    return None

  dated_backup_url = None
  timestamp = str(utils.utcnow().date())

  # The archive path for shutil.make_archive should be without an extension.
  backup_archive_path = os.path.join(
      os.path.dirname(os.path.normpath(directory)), timestamp)
  try:
    backup_archive_path = shutil.make_archive(backup_archive_path,
                                              BACKUP_ARCHIVE_FORMAT, directory)
    dated_backup_url = gcs_url_for_backup_file(
        backup_bucket_name, corpus.engine, corpus.project_qualified_target_name,
        timestamp)

    if not storage.copy_file_to(backup_archive_path, dated_backup_url):
      return None

    latest_backup_url = gcs_url_for_backup_file(
        backup_bucket_name, corpus.engine, corpus.project_qualified_target_name,
        LATEST_BACKUP_TIMESTAMP)

    if not storage.copy_blob(dated_backup_url, latest_backup_url):
      logs.log_error(
          'Failed to update latest corpus backup at "%s"' % latest_backup_url)
  except Exception as ex:
    logs.log_error(
        'backup_corpus failed: %s\n' % str(ex),
        backup_bucket_name=backup_bucket_name,
        directory=directory,
        backup_archive_path=backup_archive_path)

  finally:
    # Remove backup archive.
    shell.remove_file(backup_archive_path)

  return dated_backup_url


def gcs_url_for_backup_directory(backup_bucket_name, fuzzer_name,
                                 project_qualified_target_name):
  """Build GCS URL for corpus backup directory.

  Returns:
    A string giving the GCS URL.
  """
  return 'gs://%s/corpus/%s/%s/' % (backup_bucket_name, fuzzer_name,
                                    project_qualified_target_name)


def gcs_url_for_backup_file(backup_bucket_name, fuzzer_name,
                            project_qualified_target_name, date):
  """Build GCS URL for corpus backup file for the given date.

  Returns:
    A string giving the GCS url.
  """
  backup_dir = gcs_url_for_backup_directory(backup_bucket_name, fuzzer_name,
                                            project_qualified_target_name)
  backup_file = str(date) + os.extsep + BACKUP_ARCHIVE_FORMAT
  return '%s/%s' % (backup_dir.rstrip('/'), backup_file)


def legalize_filenames(file_paths):
  """Convert the name of every file in |file_paths| a name that is legal on
  Windows. Returns list of legally named files."""
  # TODO(metzman): Support legalizing filenames when called on trusted host, but
  # file paths exist on untrusted workers. This is fine for now since Linux is
  # the only supported platform on OSS-Fuzz and this functionality is not needed
  # in OSS-Fuzz.
  if environment.is_trusted_host():
    return file_paths

  illegal_chars = {'<', '>', ':', '\\', '|', '?', '*'}
  failed_to_move_files = []
  legally_named = []
  for file_path in file_paths:
    file_dir_path, basename = os.path.split(file_path)
    if not any(char in illegal_chars for char in basename):
      legally_named.append(file_path)
      continue

    # Hash file to get new name since it also lets us get rid of duplicates,
    # will not cause collisions for different files and makes things more
    # consistent (since libFuzzer uses hashes).
    sha1sum = utils.file_hash(file_path)
    new_file_path = os.path.join(file_dir_path, sha1sum)
    try:
      shutil.move(file_path, new_file_path)
      legally_named.append(new_file_path)
    except OSError:
      failed_to_move_files.append((file_path, new_file_path))
  if failed_to_move_files:
    logs.log_error(
        'Failed to rename files.', failed_to_move_files=failed_to_move_files)

  return legally_named


def legalize_corpus_files(directory):
  """Convert the name of every corpus file in |directory| to a name that is
  allowed on Windows."""
  # Iterate through return value of legalize_filenames to convert every
  # filename.
  files_list = shell.get_files_list(directory)
  legalize_filenames(files_list)


class GcsCorpus(object):
  """Google Cloud Storage corpus."""

  def __init__(self,
               bucket_name,
               bucket_path='/',
               log_results=True,
               _gsutil_runner=DEFAULT_GSUTIL_RUNNER):
    """Inits the GcsCorpus.

    Args:
      bucket_name: Name of the bucket for corpus synchronization.
      bucket_path: Path in the bucket where the corpus is stored.
    """
    self._bucket_name = bucket_name
    self._bucket_path = bucket_path
    self._log_results = log_results
    self._gsutil_runner = _gsutil_runner()

  @property
  def bucket_name(self):
    return self._bucket_name

  @property
  def bucket_path(self):
    return self._bucket_path

  def get_gcs_url(self, suffix=''):
    """Build corpus GCS URL for gsutil.
    Returns:
      A string giving the GCS URL.
    """
    url = 'gs://%s' % self.bucket_name + self.bucket_path + suffix
    if not url.endswith('/'):
      # Ensure that the bucket path is '/' terminated. Without this, when a
      # single file is being uploaded, it is renamed to the trailing non-/
      # terminated directory name instead.
      url += '/'

    return url

  def rsync_from_disk(self,
                      directory,
                      timeout=CORPUS_FILES_SYNC_TIMEOUT,
                      delete=True):
    """Upload local files to GCS and remove files which do not exist locally.

    Args:
      directory: Path to directory to sync from.
      timeout: Timeout for gsutil.
      delete: Whether or not to delete files on GCS that don't exist locally.

    Returns:
      A bool indicating whether or not the command succeeded.
    """
    corpus_gcs_url = self.get_gcs_url()
    legalize_corpus_files(directory)
    result = self._gsutil_runner.rsync(directory, corpus_gcs_url, timeout,
                                       delete)

    # Allow a small number of files to fail to be synced.
    return _handle_rsync_result(result, max_errors=MAX_SYNC_ERRORS)

  def rsync_to_disk(self,
                    directory,
                    timeout=CORPUS_FILES_SYNC_TIMEOUT,
                    delete=True):
    """Run gsutil to download corpus files from GCS.

    Args:
      directory: Path to directory to sync to.
      timeout: Timeout for gsutil.
      delete: Whether or not to delete files on disk that don't exist locally.

    Returns:
      A bool indicating whether or not the command succeeded.
    """
    shell.create_directory(directory, create_intermediates=True)

    corpus_gcs_url = self.get_gcs_url()
    result = self._gsutil_runner.rsync(corpus_gcs_url, directory, timeout,
                                       delete)

    # Allow a small number of files to fail to be synced.
    return _handle_rsync_result(result, max_errors=MAX_SYNC_ERRORS)

  def upload_files(self, file_paths, timeout=CORPUS_FILES_SYNC_TIMEOUT):
    """Upload files to the GCS.

    Args:
      file_paths: A sequence of file paths to upload.

    Returns:
      A bool indicating whether or not the command succeeded.
    """
    if not file_paths:
      return True

    # Get a new file_paths iterator where all files have been renamed to be
    # legal on Windows.
    file_paths = legalize_filenames(file_paths)
    gcs_url = self.get_gcs_url()
    return self._gsutil_runner.upload_files_to_url(
        file_paths, gcs_url, timeout=timeout)


class FuzzTargetCorpus(GcsCorpus):
  """Engine fuzzer (libFuzzer, AFL) specific corpus."""

  def __init__(self,
               engine,
               project_qualified_target_name,
               quarantine=False,
               log_results=True,
               include_regressions=False,
               _gsutil_runner=DEFAULT_GSUTIL_RUNNER):
    """Inits the FuzzTargetCorpus.

    Args:
      engine: The engine name. e.g. "libFuzzer".
      project_qualified_target_name: The project qualified fuzzer name. e.g.
          "libxml2_xml_read_memory_fuzzer".
      quarantine: A bool indicating whether or not this is the quarantine
          corpus.

    Raises:
      RuntimeError: If the required environment variables are not set.
    """

    # This is used to let AFL share corpora with libFuzzer.
    self._engine = os.getenv('CORPUS_FUZZER_NAME_OVERRIDE', engine)
    self._project_qualified_target_name = project_qualified_target_name

    if quarantine:
      sync_corpus_bucket_name = environment.get_value('QUARANTINE_BUCKET')
    else:
      sync_corpus_bucket_name = environment.get_value('CORPUS_BUCKET')

    if not sync_corpus_bucket_name:
      raise RuntimeError('No corpus bucket specified.')

    GcsCorpus.__init__(
        self,
        sync_corpus_bucket_name,
        '/%s/%s' % (self._engine, self._project_qualified_target_name),
        log_results=log_results,
        _gsutil_runner=_gsutil_runner,
    )

    self._regressions_corpus = GcsCorpus(
        sync_corpus_bucket_name,
        '/%s/%s%s' % (self._engine, self._project_qualified_target_name,
                      REGRESSIONS_GCS_PATH_SUFFIX),
        log_results=log_results,
        _gsutil_runner=_gsutil_runner) if include_regressions else None

  @property
  def engine(self):
    return self._engine

  @property
  def project_qualified_target_name(self):
    return self._project_qualified_target_name

  def rsync_from_disk(self,
                      directory,
                      timeout=CORPUS_FILES_SYNC_TIMEOUT,
                      delete=True):
    """Upload local files to GCS and remove files which do not exist locally.

    Overridden to have additional logging.

    Args:
      directory: Path to directory to sync to.
      timeout: Timeout for gsutil.
      delete: Whether or not to delete files on GCS that don't exist locally.

    Returns:
      A bool indicating whether or not the command succeeded.
    """
    result = GcsCorpus.rsync_from_disk(
        self, directory, timeout=timeout, delete=delete)

    num_files = _count_corpus_files(directory)
    if self._log_results:
      logs.log('%d corpus files uploaded for %s.' %
               (num_files, self._project_qualified_target_name))

    return result

  def rsync_to_disk(self,
                    directory,
                    timeout=CORPUS_FILES_SYNC_TIMEOUT,
                    delete=True):
    """Run gsutil to download corpus files from GCS.

    Overridden to have additional logging.

    Args:
      directory: Path to directory to sync to.
      timeout: Timeout for gsutil.
      delete: Whether or not to delete files on disk that don't exist locally.

    Returns:
      A bool indicating whether or not the command succeeded.
    """
    result = GcsCorpus.rsync_to_disk(
        self, directory, timeout=timeout, delete=delete)
    if not result:
      return False

    # Checkout additional regressions corpus if set and ignore the result.
    if self._regressions_corpus:
      regressions_dir = os.path.join(directory, 'regressions')
      self._regressions_corpus.rsync_to_disk(
          regressions_dir, timeout=timeout, delete=False)

    num_files = _count_corpus_files(directory)
    if self._log_results:
      logs.log('%d corpus files downloaded for %s.' %
               (num_files, self._project_qualified_target_name))

    return result

  def get_regressions_corpus_gcs_url(self):
    """Return gcs path to directory containing crash regressions."""
    return self.get_gcs_url(suffix=REGRESSIONS_GCS_PATH_SUFFIX)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common functionality for the crash stats page and the crash stats in the
  testcase detail page. This can't be with common.crash_stats because it imports
  specific libraries on appengine."""

from clusterfuzz._internal.metrics import crash_stats
from libs import helpers
from libs.query import big_query_query

# We don't allow showing more than 3 days when viewing by hours because it'd
# break the UI.
MAX_DAYS_FOR_BY_HOURS = 3


class Query(big_query_query.Query):
  """A query class for crash stats. It contains two additional fields."""

  def __init__(self):
    self.end = None
    self.days = None
    self.block = None
    self.group_by = None
    super(Query, self).__init__()

  def set_time_params(self, end, days, block):
    """Set time-related params."""
    if block == 'hour' and days > MAX_DAYS_FOR_BY_HOURS:
      raise helpers.EarlyExitException(
          ('When viewing by hours, selecting more than %d days is not allowed.'
           % MAX_DAYS_FOR_BY_HOURS), 400)

    # `hour` is inclusive. It includes the count between [hour, hour + 1). But,
    # on the UI, the end time is the last border. Therefore, we have to -1.
    self.end = end - 1
    self.days = days
    self.block = block


def get(query, group_query, offset, limit):
  """Query from BigQuery given the query object."""
  return crash_stats.get(
      end=query.end,
      days=query.days,
      block=query.block,
      group_by=query.group_by,
      where_clause=query.get_where_clause(),
      group_having_clause=group_query.get_where_clause(),
      sort_by=query.sort_by,
      offset=offset,
      limit=limit)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the base handler class."""

import unittest

import flask
from flask import request
import webtest

from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from handlers import base_handler
from libs import helpers


class JsonHandler(base_handler.Handler):
  """Render JSON response for testing."""

  def get(self):
    return self.render_json({'test': 'value'})


class HtmlHandler(base_handler.Handler):
  """Render HTML response for testing."""

  def get(self):
    return self.render('test.html', {'test': 'value'})


class ExceptionJsonHandler(base_handler.Handler):
  """Render exception in JSON response for testing."""

  def get(self):
    self.is_json = True
    raise Exception('message')


class ExceptionHtmlHandler(base_handler.Handler):
  """Render exception in HTML response for testing."""

  def get(self):
    raise Exception('unique_message')


class EarlyExceptionHandler(base_handler.Handler):
  """Render EarlyException in JSON for testing."""

  def get(self):
    self.is_json = True
    raise helpers.EarlyExitException('message', 500, [])


class AccessDeniedExceptionHandler(base_handler.Handler):
  """Render forbidden in HTML response for testing."""

  def get(self):
    raise helpers.AccessDeniedException('this_random_message')


class FlaskRedirectHandler(base_handler.Handler):
  """Redirect handler."""

  def get(self):
    redirect = request.args.get('redirect')
    return self.redirect(redirect)


class HandlerTest(unittest.TestCase):
  """Test Handler."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.config.db_config.get_value',
        'libs.form.generate_csrf_token',
        'libs.helpers.get_user_email',
    ])
    self.mock.get_value.return_value = 'contact_string'
    self.mock.generate_csrf_token.return_value = 'csrf_token'
    self.mock.get_user_email.return_value = 'test@test.com'

  def test_render_json(self):
    """Ensure it renders JSON correctly."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=JsonHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get('/')
    self.assertEqual(response.status_int, 200)
    self.assertDictEqual(response.json, {'test': 'value'})

  def test_render_early_exception(self):
    """Ensure it renders JSON response for EarlyExitException properly."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=EarlyExceptionHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get('/', expect_errors=True)
    self.assertEqual(response.status_int, 500)
    self.assertEqual(response.json['message'], 'message')
    self.assertEqual(response.json['email'], 'test@test.com')

  def test_render_json_exception(self):
    """Ensure it renders JSON exception correctly."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=ExceptionJsonHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get('/', expect_errors=True)
    self.assertEqual(response.status_int, 500)
    self.assertEqual(response.json['message'], 'message')
    self.assertEqual(response.json['email'], 'test@test.com')

  def test_render(self):
    """Ensure it gets template and render HTML correctly."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=HtmlHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get('/')
    self.assertEqual(response.status_int, 200)
    self.assertEqual(response.body, b'<html><body>value\n</body></html>')

  def test_render_html_exception(self):
    """Ensure it renders HTML exception correctly."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=ExceptionHtmlHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get('/', expect_errors=True)
    self.assertEqual(response.status_int, 500)
    self.assertRegex(response.body, b'.*unique_message.*')
    self.assertRegex(response.body, b'.*test@test.com.*')

  def test_forbidden_not_logged_in(self):
    """Ensure it renders forbidden response correctly (when not logged in)."""
    self.mock.get_user_email.return_value = None

    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=AccessDeniedExceptionHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get('/', expect_errors=True)
    self.assertEqual(response.status_int, 302)
    self.assertEqual('http://localhost/login?dest=http%3A%2F%2Flocalhost%2F',
                     response.headers['Location'])

  def test_forbidden_logged_in(self):
    """Ensure it renders forbidden response correctly (when logged in)."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=AccessDeniedExceptionHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get('/', expect_errors=True)
    self.assertEqual(response.status_int, 403)
    self.assertRegex(response.body, b'.*Access Denied.*')
    self.assertRegex(response.body, b'.*this_random_message.*')

  def test_redirect_another_page(self):
    """Test redirect to another page."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=FlaskRedirectHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get('/?redirect=%2Fanother-page')
    self.assertEqual('http://localhost/another-page',
                     response.headers['Location'])

  def test_redirect_another_domain(self):
    """Test redirect to another domain."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=FlaskRedirectHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get('/?redirect=https%3A%2F%2Fblah.com%2Ftest')
    self.assertEqual('https://blah.com/test', response.headers['Location'])

  def test_redirect_javascript(self):
    """Test redirect to a javascript url."""
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=FlaskRedirectHandler.as_view('/'))
    app = webtest.TestApp(flaskapp)
    response = app.get(
        '/?redirect=javascript%3Aalert%281%29', expect_errors=True)
    self.assertEqual(response.status_int, 403)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Install all required dependencies for running an appengine, a bot, and a
  mapreduce locally."""

from local.butler import appengine
from local.butler import common


def execute(args):
  """Install all required dependencies for running tests, the appengine,
    and the bot."""
  is_reproduce_tool_setup = args.only_reproduce
  common.install_dependencies(is_reproduce_tool_setup=is_reproduce_tool_setup)

  # App engine setup is not needed for the reproduce tool.
  if not is_reproduce_tool_setup:
    appengine.symlink_dirs()

  print('Bootstrap successfully finished.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""crash_comparer tests."""
import unittest

from clusterfuzz._internal.crash_analysis import crash_comparer


class CrashComparerTest(unittest.TestCase):
  """Tests CrashComparer."""

  def is_similar_helper(self, crash_state_1, crash_state_2, result=True):
    crash_comparer_instance = crash_comparer.CrashComparer(
        crash_state_1, crash_state_2)
    self.assertEqual(crash_comparer_instance.is_similar(), result)

  def test_is_similar_ascii(self):
    """Test is similar with ASCII."""
    crash_state_1 = (
        'start <= end ('
        'SAMP"0AA0AA00AAA0A00A0A0"@offsetInAnchor[19]) in TextIterator.cpp\n'
        'blink::TextIteratorAlgorithm<blink::EditingAlgorithm<blink::'
        'NodeTraversal> >::Te\n'
        'blink::PlainText\n')
    crash_state_2 = (
        'start <= end ('
        '#text "ABOUT US"@offsetInAnchor[0] vs. TABLE class="clash"@offsetI\n'
        'blink::TextIteratorAlgorithm<blink::EditingAlgorithm<blink::'
        'FlatTreeTraversal> >\n'
        'blink::PlainText\n')
    self.is_similar_helper(crash_state_1, crash_state_2, True)

  def test_is_similar_unicode(self):
    """Test is similar with unicode."""
    crash_state_1 = (
        'view_it != guest_view_registry_.end().Invalid GuestView created of '
        'type "V90P\n'
        'guest_view::GuestViewManager::ViewCreated\n'
        'guest_view::GuestViewMessageFilter::OnViewCreated\n')
    crash_state_2 = (
        'view_it != guest_view_registry_.end().Invalid GuestView created of '
        'type "_< i\n'
        'guest_view::GuestViewManager::ViewCreated\n'
        'guest_view::GuestViewMessageFilter::OnViewCreated\n')
    self.is_similar_helper(crash_state_1, crash_state_2, True)

  def test_is_not_similar(self):
    """Test not is similar."""
    crash_state_1 = (
        'start <= end ('
        'SAMP"0AA0AA00AAA0A00A0A0"@offsetInAnchor[19]) in TextIterator.cpp\n'
        'blink::TextIteratorAlgorithm<blink::EditingAlgorithm<blink::'
        'NodeTraversal> >::Te\n'
        'blink::PlainText\n')
    crash_state_2 = ('false in gles2_cmd_utils.cc\n'
                     'base::debug::DebugBreak\n'
                     'gpu::gles2::GLES2Util::GLFaceTargetToTextureTarget\n')
    self.is_similar_helper(crash_state_1, crash_state_2, False)

  def test_is_similar_shifted(self):
    """Test similar when frames have shifted slightly."""
    crash_state_1 = 'first\nsecond\nthird\n'
    crash_state_2 = 'second\nthird\nfourth\n'
    self.is_similar_helper(crash_state_1, crash_state_2, True)

    crash_state_2 = 'first\nthird\nfourth'
    self.is_similar_helper(crash_state_1, crash_state_2, True)

    crash_state_2 = 'first\nsecond\nfourth'
    self.is_similar_helper(crash_state_1, crash_state_2, True)

    crash_state_2 = 'first\nthird'
    self.is_similar_helper(crash_state_1, crash_state_2, True)

    crash_state_2 = 'other\nsecond\nthird'
    self.is_similar_helper(crash_state_1, crash_state_2, True)

  def test_only_one_frame_matching(self):
    """Test not similar when 1/3 frames match."""
    crash_state_1 = 'first\nsecond\nthird\n'
    crash_state_2 = 'second\nfourth\nfifth\n'
    self.is_similar_helper(crash_state_1, crash_state_2, False)

    crash_state_2 = 'first\nfourth\nfifth\n'
    self.is_similar_helper(crash_state_1, crash_state_2, False)

  def test_is_same_frames_wrong_order(self):
    """Test not similar when some frames match but the order doesn't match."""
    crash_state_1 = 'first\nsecond\nthird\n'
    crash_state_2 = 'second\nfirst\nfourth\n'
    self.is_similar_helper(crash_state_1, crash_state_2, False)

    crash_state_2 = 'third\nsecond\nfirst\n'
    self.is_similar_helper(crash_state_1, crash_state_2, False)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for fetching build info from OmahaProxy."""

import json
import re

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

BUILD_INFO_PATTERN = ('([a-z]+),([a-z]+),([0-9.]+),'
                      '[^,]*,[^,]*,[^,]*,[^,]*,[^,]*,'
                      '([0-9a-f]+),.*')
BUILD_INFO_URL = 'https://omahaproxy.appspot.com/all?csv=1'
BUILD_INFO_URL_CD = ('https://chromiumdash.appspot.com/fetch_releases?'
                     'num=1&platform={platform}')


class BuildInfo(object):
  """BuildInfo holds build metadata pulled from OmahaProxy."""

  def __init__(self, platform, build_type, version, revision):
    self.platform = platform
    self.build_type = build_type
    self.version = version
    self.revision = revision


def _convert_platform_to_omahaproxy_platform(platform):
  """Converts platform to omahaproxy platform for use in
  get_production_builds_info."""
  platform_lower = platform.lower()
  if platform_lower == 'windows':
    return 'win'
  return platform_lower


def _convert_platform_to_chromiumdash_platform(platform):
  """Converts platform to Chromium Dash platform.
  Note that Windows in Chromium Dash is win64 and we only want win32."""
  platform_lower = platform.lower()
  if platform_lower == 'windows':
    return 'Win32'
  return platform_lower.capitalize()


def _fetch_releases_from_chromiumdash(platform, channel=None):
  """Makes a Call to chromiumdash's fetch_releases api,
  and returns its json array response."""
  chromiumdash_platform = _convert_platform_to_chromiumdash_platform(platform)
  query_url = BUILD_INFO_URL_CD.format(platform=chromiumdash_platform)
  if channel:
    query_url = query_url + '&channel=' + channel

  build_info = utils.fetch_url(query_url)
  if not build_info:
    logs.log_error('Failed to fetch build info from %s' % query_url)
    return []

  try:
    build_info_json = json.loads(build_info)
    if not build_info_json:
      logs.log_error('Empty response from %s' % query_url)
      return []
  except Exception:
    logs.log_error('Malformed response from %s' % query_url)
    return []

  return build_info_json


def get_production_builds_info(platform):
  """Gets the build information for production builds.

  Omits platforms containing digits, namely, win64.
  Omits channels containing underscore, namely, canary_asan.
  Platform is e.g. ANDROID, LINUX, MAC, WIN.
  """
  builds_metadata = []
  omahaproxy_platform = _convert_platform_to_omahaproxy_platform(platform)

  build_info = utils.fetch_url(BUILD_INFO_URL)
  if not build_info:
    logs.log_error('Failed to fetch build info from %s' % BUILD_INFO_URL)
    return []

  for line in build_info.splitlines():
    match = re.match(BUILD_INFO_PATTERN, line)
    if not match:
      continue

    platform_type = match.group(1)
    if platform_type != omahaproxy_platform:
      continue

    build_type = match.group(2)
    version = match.group(3)
    revision = match.group(4)
    builds_metadata.append(BuildInfo(platform, build_type, version, revision))

  return builds_metadata


def get_production_builds_info_from_cd(platform):
  """Gets the build information from Chromium Dash for production builds.

  Omits platforms containing digits, namely, win64.
  Omits channels containing underscore, namely, canary_asan.
  Platform is e.g. ANDROID, LINUX, MAC, WINDOWS.
  """
  builds_metadata = []
  build_info_json = _fetch_releases_from_chromiumdash(platform)
  for info in build_info_json:
    build_type = info['channel'].lower()
    if build_type == 'extended':
      build_type = 'extended_stable'

    version = info['version']
    revision = info['hashes']['chromium']
    builds_metadata.append(BuildInfo(platform, build_type, version, revision))

  # Hack: pretend Windows extended stable info to be Linux extended stable info.
  # Because Linux doesn't have extended stable channel.
  if platform.lower() == 'linux':
    es_info = _fetch_releases_from_chromiumdash(
        'WINDOWS', channel='Extended')[0]
    builds_metadata.append(
        BuildInfo(platform, 'extended_stable', es_info['version'],
                  es_info['hashes']['chromium']))

  return builds_metadata


def get_release_milestone(build_type, platform):
  """Return milestone for a particular release."""
  if build_type == 'head':
    actual_build_type = 'canary'
  else:
    actual_build_type = build_type

  builds_metadata = get_production_builds_info_from_cd(platform)
  for build_metadata in builds_metadata:
    if build_metadata.build_type == actual_build_type:
      version_parts = build_metadata.version.split('.')
      milestone = version_parts[0]
      if milestone and milestone.isdigit():
        return int(milestone)

  if actual_build_type == 'canary':
    # If there is no canary for that platform, just return canary from windows.
    return get_release_milestone('canary', 'windows')

  return None


def get_build_to_revision_mappings(platform=None):
  """Gets the build information."""
  if not platform:
    platform = environment.platform()

  result = {}
  build_info_json = _fetch_releases_from_chromiumdash(platform)

  for info in build_info_json:
    build_type = info['channel'].lower()
    if build_type == 'extended':
      build_type = 'extended_stable'

    version = info['version']
    revision = str(info['chromium_main_branch_position'])
    result[build_type] = {'revision': revision, 'version': version}

  # Hack: pretend Windows extended stable info to be Linux extended stable info.
  # Because Linux doesn't have extended stable channel.
  if platform.lower() == 'linux':
    es_info = _fetch_releases_from_chromiumdash(
        'WINDOWS', channel='Extended')[0]
    result['extended_stable'] = {
        'revision': str(es_info['chromium_main_branch_position']),
        'version': es_info['version']
    }

  return result



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GFT engine interface."""

import os
import re

from clusterfuzz._internal.system import new_process
from clusterfuzz.fuzz import engine

_CRASH_REGEX = re.compile(r'.*Reproducer file written to:\s*(.*)$')


class GoogleFuzzTestError(Exception):
  """Base exception class."""


def _get_reproducer_path(line):
  """Get the reproducer path, if any."""
  crash_match = _CRASH_REGEX.match(line)
  if not crash_match:
    return None

  return crash_match.group(1)


class GoogleFuzzTestEngine(engine.Engine):
  """GFT engine implementation."""

  @property
  def name(self):
    return 'googlefuzztest'

  def prepare(self, corpus_dir, target_path, build_dir):  # pylint: disable=unused-argument
    """Prepare for a fuzzing session, by generating options. Returns a
    FuzzOptions object.

    Args:
      corpus_dir: The main corpus directory.
      target_path: Path to the target.
      build_dir: Path to the build directory.

    Returns:
      A FuzzOptions object.
    """
    os.chmod(target_path, 0o775)
    return engine.FuzzOptions(corpus_dir, [], {})

  def fuzz(self, target_path, options, reproducers_dir, max_time):
    """Run a fuzz session.

    Args:
      target_path: Path to the target.
      options: The FuzzOptions object returned by prepare().
      reproducers_dir: The directory to put reproducers in when crashes
          are found.
      max_time: Maximum allowed time for the fuzzing to run.

   Returns:
      A FuzzResult object.
    """
    del options  # Unused.
    runner = new_process.UnicodeProcessRunner(target_path)

    fuzz_result = runner.run_and_wait(
        timeout=max_time,
        extra_env={
            'FUZZTEST_REPRODUCERS_OUT_DIR': reproducers_dir,
        })
    log_lines = fuzz_result.output.splitlines()

    crashes = []
    for line in log_lines:
      reproducer_path = _get_reproducer_path(line)
      if reproducer_path:
        crashes.append(
            engine.Crash(
                reproducer_path,
                fuzz_result.output,
                reproduce_args=[],
                crash_time=int(fuzz_result.time_executed)))
        continue

    # TODO(ochang): Implement stats parsing.
    stats = {}
    return engine.FuzzResult(fuzz_result.output, fuzz_result.command, crashes,
                             stats, fuzz_result.time_executed)

  def reproduce(self, target_path, input_path, arguments, max_time):  # pylint: disable=unused-argument
    """Reproduce a crash given an input.

    Args:
      target_path: Path to the target.
      input_path: Path to the reproducer input.
      arguments: Additional arguments needed for reproduction.
      max_time: Maximum allowed time for the reproduction.

    Returns:
      A ReproduceResult.
    """
    os.chmod(target_path, 0o775)
    runner = new_process.UnicodeProcessRunner(target_path)
    result = runner.run_and_wait(
        timeout=max_time, extra_env={'FUZZTEST_REPLAY': input_path})

    return engine.ReproduceResult(result.command, result.return_code,
                                  result.time_executed, result.output)

  def minimize_corpus(self, target_path, arguments, input_dirs, output_dir,
                      reproducers_dir, max_time):
    """Optional (but recommended): run corpus minimization.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for corpus minimization.
      input_dirs: Input corpora.
      output_dir: Output directory to place minimized corpus.
      reproducers_dir: The directory to put reproducers in when crashes are
          found.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A FuzzResult object.

    Raises:
      TimeoutError: If the corpus minimization exceeds max_time.
      Error: If the merge failed in some other way.
    """
    raise NotImplementedError

  def minimize_testcase(self, target_path, arguments, input_path, output_path,
                        max_time):
    """Optional (but recommended): Minimize a testcase.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for testcase minimization.
      input_path: Path to the reproducer input.
      output_path: Path to the minimized output.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A ReproduceResult.

    Raises:
      TimeoutError: If the testcase minimization exceeds max_time.
    """
    raise NotImplementedError

  def cleanse(self, target_path, arguments, input_path, output_path, max_time):
    """Optional (but recommended): Cleanse a testcase.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for testcase cleanse.
      input_path: Path to the reproducer input.
      output_path: Path to the cleansed output.
      max_time: Maximum allowed time for the cleanse.

    Returns:
      A ReproduceResult.

    Raises:
      TimeoutError: If the cleanse exceeds max_time.
    """
    raise NotImplementedError



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common constants for LKL."""

import re

LINUX_KERNEL_MODULE_STACK_TRACE = 'Linux Kernel Library Stack Trace:'

#hid-fuzzer: lib/posix-host.c:401: void panic(void): Assertion `0' failed.
LINUX_KERNEL_LIBRARY_ASSERT_REGEX = re.compile(
    r'([^:]+): lib/posix-host\.c:\d+: void panic\(void\): Assertion .*')

# Linux version 5.4.58+-ab6926695 where 6926695 is the build id.
# Unlike in a normal linux version string, we do not know the build hash.
LINUX_VERSION_REGEX_LKL = re.compile(r'Linux version .+-(ab([0-9a-f]+)\s)')

# This is the prefix in the repo.prop for the kernel for all
# lkl fuzzers.
LKL_REPO_KERNEL_PREFIX = 'kernel/private/lkl'
LKL_BUILD_TARGET = 'kernel_kasan.lkl_fuzzers'



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fuzzer stats."""
# pylint: disable=protected-access
import ast
import datetime
import json
import os
import unittest

import flask
import mock
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers import fuzzer_stats
from libs import helpers

DATA_DIRECTORY = os.path.join(os.path.dirname(__file__), 'fuzzer_stats_data')


def _read_data_file(data_file):
  """Helper function to read the contents of a data file."""
  with open(os.path.join(DATA_DIRECTORY, data_file)) as handle:
    return handle.read()


def _mock_query(fuzzer, jobs, group_by, date_start, date_end):
  return json.loads(
      _read_data_file('%s_%s_%s_%s_%s.txt' % (fuzzer, '_'.join(jobs)
                                              if jobs else None, group_by,
                                              date_start, date_end)))


@test_utils.with_cloud_emulators('datastore')
class TestBuildResults(unittest.TestCase):
  """Tests for fuzzer_stats build_results and wrappers."""

  def setUp(self):
    test_helpers.patch_environ(self)
    self.maxDiff = None  # pylint: disable=invalid-name

    data_types.Fuzzer(
        name='testFuzzer',
        stats_columns=('sum(t.blah) as blah, custom(j.new_crashes) '
                       'as new_crashes, _EDGE_COV as edge_coverage, '
                       '_FUNC_COV as func_coverage, '
                       '_CORPUS_SIZE as corpus_size, '
                       '_CORPUS_BACKUP as corpus_backup, '
                       '_QUARANTINE_SIZE as quarantine_size, '
                       '_COV_REPORT as coverage_report, '
                       '_FUZZER_RUN_LOGS as fuzzer_logs,'
                       '_PERFORMANCE_REPORT as performance_report'),
        stats_column_descriptions=(
            'blah: "blah description"\n'
            'func_coverage: "func coverage description"\n')).put()

    data_types.Fuzzer(
        name='testFuzzer2',
        stats_columns=('sum(t.blah) as blah, custom(j.new_crashes) '
                       'as new_crashes, _EDGE_COV as edge_coverage, '
                       '_FUNC_COV as func_coverage, '
                       '_CORPUS_SIZE as corpus_size, '
                       '_CORPUS_BACKUP as corpus_backup, '
                       '_QUARANTINE_SIZE as quarantine_size, '
                       '_COV_REPORT as coverage_report, '
                       '_FUZZER_RUN_LOGS as fuzzer_logs,'
                       '_PERFORMANCE_REPORT as performance_report'),
        stats_column_descriptions=(
            'blah: "blah description"\n'
            'func_coverage: "func coverage description"\n')).put()

    data_types.Job(
        name='job', environment_string='FUZZ_LOGS_BUCKET = bucket').put()

    now = datetime.datetime.utcnow()

    data_types.FuzzTarget(
        engine='testFuzzer', project='test-project', binary='1_fuzzer').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='testFuzzer_1_fuzzer', job='job', last_run=now).put()

    data_types.FuzzTarget(
        engine='testFuzzer', project='test-project', binary='2_fuzzer').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='testFuzzer_2_fuzzer', job='job', last_run=now).put()

    data_types.FuzzTarget(
        engine='testFuzzer', project='test-project', binary='3_fuzzer').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='testFuzzer_3_fuzzer', job='job', last_run=now).put()

    data_types.FuzzTarget(
        engine='testFuzzer2', project='test-project', binary='1_fuzzer').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='testFuzzer2_1_fuzzer', job='job', last_run=now).put()

    cov_info = data_types.CoverageInformation(
        fuzzer='2_fuzzer', date=datetime.date(2016, 10, 19))
    cov_info.edges_covered = 11
    cov_info.edges_total = 30
    cov_info.functions_covered = 10
    cov_info.functions_total = 15
    cov_info.html_report_url = 'https://report_for_2_fuzzer/20161019'
    cov_info.corpus_size_units = 20
    cov_info.corpus_size_bytes = 200
    cov_info.quarantine_size_units = 0
    cov_info.quarantine_size_bytes = 0
    cov_info.corpus_location = 'gs://corpus'
    cov_info.corpus_backup_location = 'gs://corpus-backup/file.zip'
    cov_info.quarantine_location = 'gs://quarantine'
    cov_info.put()

    cov_info = data_types.CoverageInformation(
        fuzzer='2_fuzzer', date=datetime.date(2016, 10, 21))
    cov_info.edges_covered = 15
    cov_info.edges_total = 30
    cov_info.functions_covered = 11
    cov_info.functions_total = 15
    cov_info.html_report_url = 'https://report_for_2_fuzzer/20161021'
    cov_info.corpus_size_units = 40
    cov_info.corpus_size_bytes = 400
    cov_info.quarantine_size_units = 8
    cov_info.quarantine_size_bytes = 80
    cov_info.corpus_location = 'gs://corpus'
    cov_info.corpus_backup_location = 'gs://corpus-backup/file.zip'
    cov_info.quarantine_location = 'gs://quarantine'
    cov_info.put()

    cov_info = data_types.CoverageInformation(
        fuzzer='1_fuzzer', date=datetime.date(2016, 10, 20))
    cov_info.edges_covered = 17
    cov_info.edges_total = 38
    cov_info.functions_covered = 12
    cov_info.functions_total = 19
    cov_info.html_report_url = 'https://report_for_1_fuzzer/20161020'
    cov_info.corpus_size_units = 47
    cov_info.corpus_size_bytes = 480
    cov_info.quarantine_size_units = 3
    cov_info.quarantine_size_bytes = 8
    cov_info.corpus_location = 'gs://corpus'
    cov_info.corpus_backup_location = 'gs://corpus-backup/file.zip'
    cov_info.quarantine_location = 'gs://quarantine'
    cov_info.put()

    self.client = mock.Mock(spec_set=big_query.Client)
    test_helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.big_query.Client',
    ])
    self.mock.Client.return_value = self.client

  def test_build_invalid_params(self):
    """Tests build_results with invalid/missing params."""
    with self.assertRaises(helpers.EarlyExitException):
      fuzzer_stats.build_results('', '', '', '', '')

  def test_build_by_fuzzer(self):
    """Tests basic build_results with valid parameters (group by fuzzer)."""
    build_args = [
        'testFuzzer', ['job'], 'by-fuzzer', '2016-10-20', '2016-10-21'
    ]
    self.client.raw_query.return_value = _mock_query(*build_args)
    result = fuzzer_stats.build_results(*build_args)

    self.assertDictEqual(
        ast.literal_eval(_read_data_file('by_fuzzer_expected.txt')), result)

  def test_build_by_day(self):
    """Tests basic build_results with valid parameters (group by day)."""
    build_args = [
        'testFuzzer_2_fuzzer', ['job'], 'by-day', '2016-10-19', '2016-10-21'
    ]
    self.client.raw_query.return_value = _mock_query(*build_args)
    result = fuzzer_stats.build_results(*build_args)

    self.assertDictEqual(
        ast.literal_eval(_read_data_file('by_day_expected.txt')), result)

  def test_build_by_job(self):
    """Tests basic build_results with valid parameters (group by job)."""
    build_args = [
        'testFuzzer_1_fuzzer', None, 'by-job', '2016-10-20', '2016-10-21'
    ]
    self.client.raw_query.return_value = _mock_query(*build_args)
    result = fuzzer_stats.build_results(*build_args)

    self.assertDictEqual(
        ast.literal_eval(_read_data_file('by_job_expected.txt')), result)

  def test_build_by_time(self):
    """Tests basic build_results with valid parameters (group by time)."""
    build_args = [
        'testFuzzer_2_fuzzer', ['job'], 'by-time', '2016-10-19', '2016-10-21'
    ]
    self.client.raw_query.return_value = _mock_query(*build_args)
    result = fuzzer_stats.build_results(*build_args)

    self.assertDictEqual(
        ast.literal_eval(_read_data_file('by_time_expected.txt')), result)

  def test_build_none_results(self):
    """Tests basic build_results, with bigquery returning None for some
    columns."""
    build_args = ['testFuzzer', None, 'by-fuzzer', '2016-10-20', '2016-10-22']
    self.client.raw_query.return_value = _mock_query(*build_args)
    result = fuzzer_stats.build_results(*build_args)

    self.assertDictEqual(
        ast.literal_eval(_read_data_file('by_fuzzer_expected_None.txt')),
        result)

  def test_build_old_results(self):
    """Tests build_results is caching old results appropriately."""
    test_helpers.patch(self, [
        'handlers.fuzzer_stats._build_old_results',
        'handlers.fuzzer_stats._build_todays_results',
        'clusterfuzz._internal.base.utils.utcnow',
    ])

    self.mock._build_old_results.return_value = None
    self.mock._build_todays_results.return_value = None
    self.mock.utcnow.return_value = datetime.datetime(2018, 5, 4)

    date_yesterday = '2018-05-03'
    build_args = [
        'testFuzzer_2_fuzzer', ['job'], 'by-time', date_yesterday,
        date_yesterday
    ]

    # Test that results from yesterday are cached for a full day.
    fuzzer_stats.build_results(*build_args)
    self.mock._build_old_results.assert_called_with(*build_args)
    self.assertEqual(0, self.mock._build_todays_results.call_count)

  def test_build_today_results(self):
    """Tests build_results is caching results from today appropriately."""
    test_helpers.patch(self, [
        'handlers.fuzzer_stats._build_old_results',
        'handlers.fuzzer_stats._build_todays_results',
        'clusterfuzz._internal.base.utils.utcnow',
    ])

    self.mock._build_old_results.return_value = None
    self.mock._build_todays_results.return_value = None
    self.mock.utcnow.return_value = datetime.datetime(2018, 5, 4)

    date_today = '2018-05-04'
    build_args = [
        'testFuzzer_2_fuzzer', ['job'], 'by-time', date_today, date_today
    ]

    # Test that results from yesterday are cached for a full day.
    fuzzer_stats.build_results(*build_args)
    self.mock._build_todays_results.assert_called_with(*build_args)
    self.assertEqual(0, self.mock._build_old_results.call_count)


@test_utils.with_cloud_emulators('datastore')
class TestPermissions(unittest.TestCase):
  """Permissions tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.auth.get_current_user',
        'libs.auth.is_current_user_admin',
        'handlers.fuzzer_stats.build_results',
    ])

    self.mock.build_results.return_value = json.dumps({})

    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/fuzzer-stats/load',
        view_func=fuzzer_stats.LoadHandler.as_view('/fuzzer-stats/load'))
    self.app = webtest.TestApp(flaskapp)

    data_types.ExternalUserPermission(
        email='test@user.com',
        entity_kind=data_types.PermissionEntityKind.JOB,
        entity_name='job1',
        is_prefix=False,
        auto_cc=data_types.AutoCCType.ALL).put()

    data_types.ExternalUserPermission(
        email='test@user.com',
        entity_kind=data_types.PermissionEntityKind.JOB,
        entity_name='job2',
        is_prefix=False,
        auto_cc=data_types.AutoCCType.ALL).put()

    data_types.Job(name='job1').put()
    data_types.Job(name='job2').put()
    data_types.Job(name='job3').put()

  def test_internal_user_with_job(self):
    """Test internal user access (with job)."""
    self.mock.is_current_user_admin.return_value = True
    response = self.app.post_json(
        '/fuzzer-stats/load', {
            'fuzzer': 'fuzzer',
            'job': 'job1',
            'group_by': 'by-fuzzer',
            'date_start': '2017-01-01',
            'date_end': '2017-01-07',
        })
    self.assertEqual(200, response.status_int)
    self.mock.build_results.assert_called_with('fuzzer', ['job1'], 'by-fuzzer',
                                               '2017-01-01', '2017-01-07')

  def test_internal_user_without_job(self):
    """Test internal user access (with job)."""
    self.mock.is_current_user_admin.return_value = True
    response = self.app.post_json(
        '/fuzzer-stats/load', {
            'fuzzer': 'fuzzer',
            'group_by': 'by-fuzzer',
            'date_start': '2017-01-01',
            'date_end': '2017-01-07',
        })
    self.assertEqual(200, response.status_int)
    self.mock.build_results.assert_called_with('fuzzer', None, 'by-fuzzer',
                                               '2017-01-01', '2017-01-07')

  def test_external_user_with_job(self):
    """Test external user access (job specified)."""
    self.mock.is_current_user_admin.return_value = False
    self.mock.get_current_user().email = 'test@user.com'

    response = self.app.post_json(
        '/fuzzer-stats/load', {
            'fuzzer': 'fuzzer',
            'job': 'job1',
            'group_by': 'by-fuzzer',
            'date_start': '2017-01-01',
            'date_end': '2017-01-07',
        })
    self.assertEqual(200, response.status_int)

    response = self.app.post_json(
        '/fuzzer-stats/load', {
            'fuzzer': 'fuzzer',
            'job': 'job2',
            'group_by': 'by-fuzzer',
            'date_start': '2017-01-01',
            'date_end': '2017-01-07',
        })
    self.assertEqual(200, response.status_int)

    response = self.app.post_json(
        '/fuzzer-stats/load', {
            'fuzzer': 'fuzzer',
            'job': 'job3',
            'group_by': 'by-fuzzer',
            'date_start': '2017-01-01',
            'date_end': '2017-01-07',
        },
        expect_errors=True)
    self.assertEqual(403, response.status_int)

  def test_external_user_without_job(self):
    """Test external user access (no job specified)."""
    self.mock.is_current_user_admin.return_value = False
    self.mock.get_current_user().email = 'test@user.com'

    response = self.app.post_json(
        '/fuzzer-stats/load', {
            'fuzzer': 'fuzzer',
            'group_by': 'by-fuzzer',
            'date_start': '2017-01-01',
            'date_end': '2017-01-07',
        })
    self.assertEqual(200, response.status_int)
    self.mock.build_results.assert_called_with(
        'fuzzer', ['job1', 'job2'], 'by-fuzzer', '2017-01-01', '2017-01-07')


class TestGetDate(unittest.TestCase):
  """Test for fuzzer_stats._get_date."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
    ])

  def test_get_date_empty_date(self):
    """Test _get_date correctly replaces an empty date."""
    self.mock.utcnow.return_value = datetime.datetime(2018, 5, 4)
    # Test that when given '' and 1, it returns the string representing
    # the date one day ago.
    self.assertEqual(fuzzer_stats._get_date('', 1), '2018-05-03')

  def test_get_date_non_empty_date(self):
    """Test _get_date just returns a non-empty date."""
    date = '2018-5-1'
    self.assertEqual(fuzzer_stats._get_date(date, 1), date)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for handling archives."""

import lzma
import os
import tarfile
import zipfile

from clusterfuzz._internal.metrics import logs

FILE_ATTRIBUTE = '0o10'
SYMLINK_ATTRIBUTE = '0o12'

# File extensions for archive files.
ZIP_FILE_EXTENSIONS = ['.zip']
TAR_FILE_EXTENSIONS = [
    '.tar', '.tar.bz2', '.tar.gz', '.tb2', '.tbz', '.tbz2', '.tgz'
]
LZMA_FILE_EXTENSIONS = ['.tar.lzma', '.tar.xz']

ARCHIVE_FILE_EXTENSIONS = (
    ZIP_FILE_EXTENSIONS + TAR_FILE_EXTENSIONS + LZMA_FILE_EXTENSIONS)


class ArchiveType(object):
  """Type of the archive."""
  UNKNOWN = 0
  ZIP = 1
  TAR = 2
  TAR_LZMA = 3


class ArchiveFile(object):
  """File in an archive."""

  def __init__(self, name, size, handle):
    self.name = name
    self.size = size
    self.handle = handle


def iterator(archive_path,
             archive_obj=None,
             file_match_callback=None,
             should_extract=True):
  """Return an iterator for files in an archive. Extracts files if
  |should_extract| is True."""
  archive_type = get_archive_type(archive_path)

  if not file_match_callback:
    file_match_callback = lambda _: True

  def maybe_extract(extract_func, info):
    """Returns an extracted file or None if it is not supposed to be extracted.
    """
    if should_extract:
      return extract_func(info)
    return None

  if archive_type == ArchiveType.ZIP:
    try:
      with zipfile.ZipFile(archive_obj or archive_path) as zip_file:
        for info in zip_file.infolist():
          if not file_match_callback(info.filename):
            continue

          yield ArchiveFile(info.filename, info.file_size,
                            maybe_extract(zip_file.open, info))

    except (zipfile.BadZipfile, zipfile.LargeZipFile):
      logs.log_error('Bad zip file %s.' % archive_path)

  elif archive_type == ArchiveType.TAR:
    try:
      if archive_obj:
        tar_file = tarfile.open(fileobj=archive_obj)
      else:
        tar_file = tarfile.open(archive_path)

      for info in tar_file.getmembers():
        if not file_match_callback(info.name):
          continue

        yield ArchiveFile(info.name, info.size,
                          maybe_extract(tar_file.extractfile, info))
      tar_file.close()
    except tarfile.TarError:
      logs.log_error('Bad tar file %s.' % archive_path)

  elif archive_type == ArchiveType.TAR_LZMA:
    assert archive_obj is None, "LZMAFile doesn't support opening file handles."
    try:
      with lzma.LZMAFile(archive_path) as lzma_file, \
            tarfile.open(fileobj=lzma_file) as tar_file:

        error_filepaths = []
        for info in tar_file.getmembers():
          if not file_match_callback(info.name):
            continue

          try:
            yield ArchiveFile(info.name, info.size,
                              maybe_extract(tar_file.extractfile, info))

          except KeyError:  # Handle broken links gracefully.
            error_filepaths.append(info.name)
            yield ArchiveFile(info.name, info.size, None)

        if error_filepaths:
          logs.log_warn(
              'Check archive %s for broken links.' % archive_path,
              error_filepaths=error_filepaths)

    except (lzma.LZMAError, tarfile.TarError):
      logs.log_error('Bad lzma file %s.' % archive_path)

  else:
    logs.log_error('Unsupported compression type for file %s.' % archive_path)


def extracted_size(archive_path, archive_obj=None, file_match_callback=None):
  """Return the total extracted size of the archive."""
  return sum(
      f.size for f in iterator(
          archive_path, archive_obj, file_match_callback, should_extract=False))


def get_archive_type(archive_path):
  """Get the type of the archive."""

  def has_extension(extensions):
    """Returns True if |archive_path| endswith an extension in |extensions|."""
    for extension in extensions:
      if archive_path.endswith(extension):
        return True
    return False

  if has_extension(ZIP_FILE_EXTENSIONS):
    return ArchiveType.ZIP

  if has_extension(TAR_FILE_EXTENSIONS):
    return ArchiveType.TAR

  if has_extension(LZMA_FILE_EXTENSIONS):
    return ArchiveType.TAR_LZMA

  return ArchiveType.UNKNOWN


def get_file_list(archive_path, archive_obj=None, file_match_callback=None):
  """List all files in an archive."""
  return [
      f.name for f in iterator(
          archive_path, archive_obj, file_match_callback, should_extract=False)
  ]


def get_first_file_matching(search_string, archive_obj, archive_path):
  """Returns the first file with a name matching search string in archive."""
  for current_file in get_file_list(archive_path, archive_obj):
    # Exclude MAC resouce forks.
    if current_file.startswith('__MACOSX/'):
      continue

    if search_string in current_file:
      return current_file

  return None


def is_archive(filename):
  """Return true if the file is an archive."""
  return get_archive_type(filename) != ArchiveType.UNKNOWN


def unpack(archive_path,
           output_directory,
           trusted=False,
           file_match_callback=None):
  """Extracts an archive into the target directory."""
  if not os.path.exists(archive_path):
    logs.log_error('Archive %s not found.' % archive_path)
    return False

  # If the output directory is a symlink, get its actual path since we will be
  # doing directory traversal checks later when unpacking the archive.
  output_directory = os.path.realpath(output_directory)

  archive_filename = os.path.basename(archive_path)
  error_occurred = False

  # Choose to unpack all files or ones matching a particular regex.
  file_list = get_file_list(
      archive_path, file_match_callback=file_match_callback)

  archive_file_unpack_count = 0
  archive_file_total_count = len(file_list)

  # If the archive is not trusted, do file path checks to make
  # sure this archive is safe and is not attempting to do path
  # traversals.
  if not trusted:
    for filename in file_list:
      absolute_file_path = os.path.join(output_directory,
                                        os.path.normpath(filename))
      real_file_path = os.path.realpath(absolute_file_path)

      if real_file_path == output_directory:
        # Workaround for https://bugs.python.org/issue28488.
        # Ignore directories named '.'.
        continue

      if real_file_path != absolute_file_path:
        logs.log_error(
            'Directory traversal attempted while unpacking archive %s '
            '(file path=%s, actual file path=%s). Aborting.' %
            (archive_path, absolute_file_path, real_file_path))
        return False

  archive_type = get_archive_type(archive_filename)

  # Extract based on file's extension.
  if archive_type == ArchiveType.ZIP:
    zip_file_handle = open(archive_path, 'rb')
    zip_archive = zipfile.ZipFile(zip_file_handle)

    for filename in file_list:
      try:
        extracted_path = zip_archive.extract(filename, output_directory)

        # Preserve permissions for regular files. 640 is the default
        # permission for extract. If we need execute permission, we need
        # to chmod it explicitly. Also, get rid of suid bit for security
        # reasons.
        external_attr = zip_archive.getinfo(filename).external_attr >> 16
        if oct(external_attr).startswith(FILE_ATTRIBUTE):
          old_mode = external_attr & 0o7777
          new_mode = external_attr & 0o777
          new_mode |= 0o440
          needs_execute_permission = external_attr & 100

          if new_mode != old_mode or needs_execute_permission:
            # Default extract condition is 640 which is safe.
            # |new_mode| might have read+write+execute bit for
            # others, so remove those.
            new_mode &= 0o770

            os.chmod(extracted_path, new_mode)

        # Create symlink if needed (only on unix platforms).
        if (trusted and hasattr(os, 'symlink') and
            oct(external_attr).startswith(SYMLINK_ATTRIBUTE)):
          symlink_source = zip_archive.read(filename)
          if os.path.exists(extracted_path):
            os.remove(extracted_path)
          os.symlink(symlink_source, extracted_path)

        # Keep heartbeat happy by updating with our progress.
        archive_file_unpack_count += 1
        if archive_file_unpack_count % 1000 == 0:
          logs.log('Unpacked %d/%d.' % (archive_file_unpack_count,
                                        archive_file_total_count))

      except:
        # In case of errors, we try to extract whatever we can without errors.
        error_occurred = True
        continue

    logs.log('Unpacked %d/%d.' % (archive_file_unpack_count,
                                  archive_file_total_count))
    zip_archive.close()
    zip_file_handle.close()

    if error_occurred:
      logs.log_error(
          'Failed to extract everything from archive %s.' % archive_filename)

  elif archive_type in (ArchiveType.TAR, ArchiveType.TAR_LZMA):
    if archive_type == ArchiveType.TAR_LZMA:
      lzma_file = lzma.LZMAFile(archive_path)
      tar_archive = tarfile.open(fileobj=lzma_file)
    else:
      tar_archive = tarfile.open(archive_path)

    try:
      tar_archive.extractall(path=output_directory)
    except:
      # In case of errors, we try to extract whatever we can without errors.
      error_occurred = True
      logs.log_error(
          'Failed to extract everything from archive %s, trying one at a time.'
          % archive_filename)
      for filename in file_list:
        try:
          tar_archive.extract(filename, output_directory)
        except:
          continue

        # Keep heartbeat happy by updating with our progress.
        archive_file_unpack_count += 1
        if archive_file_unpack_count % 1000 == 0:
          logs.log('Unpacked %d/%d.' % (archive_file_unpack_count,
                                        archive_file_total_count))

      logs.log('Unpacked %d/%d.' % (archive_file_unpack_count,
                                    archive_file_total_count))

    tar_archive.close()
    if archive_type == ArchiveType.TAR_LZMA:
      lzma_file.close()

  else:
    logs.log_error(
        'Unsupported compression type for file %s.' % archive_filename)
    return False

  return not error_occurred



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for managing issue tracker instance."""

from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from libs import request_cache
from libs.issue_management import issue_tracker_policy
from libs.issue_management import jira
from libs.issue_management import monorail

_ISSUE_TRACKER_CACHE_CAPACITY = 8
_ISSUE_TRACKER_CONSTRUCTORS = {
    'monorail': monorail.get_issue_tracker,
    'jira': jira.get_issue_tracker
}


def register_issue_tracker(tracker_type, constructor):
  """Register an issue tracker implementation."""
  if tracker_type in _ISSUE_TRACKER_CONSTRUCTORS:
    raise ValueError(
        'Tracker type {type} is already registered.'.format(type=tracker_type))
  _ISSUE_TRACKER_CONSTRUCTORS[tracker_type] = constructor


def _get_issue_tracker_project_name(testcase=None):
  """Return issue tracker project name given a testcase or default."""
  from clusterfuzz._internal.datastore import data_handler
  job_type = testcase.job_type if testcase else None
  return data_handler.get_issue_tracker_name(job_type)


@request_cache.wrap(capacity=_ISSUE_TRACKER_CACHE_CAPACITY)
def get_issue_tracker(project_name=None):
  """Get the issue tracker with the given type and name."""
  issue_tracker_config = local_config.IssueTrackerConfig()
  if not project_name:
    from clusterfuzz._internal.datastore import data_handler
    project_name = data_handler.get_issue_tracker_name()

  issue_project_config = issue_tracker_config.get(project_name)
  if not issue_project_config:
    raise ValueError('Issue tracker for {} does not exist'.format(project_name))

  constructor = _ISSUE_TRACKER_CONSTRUCTORS.get(issue_project_config['type'])
  if not constructor:
    raise ValueError('Invalid issue tracker type: ' +
                     issue_project_config['type'])

  return constructor(project_name, issue_project_config)


def get_issue_tracker_for_testcase(testcase):
  """Get the issue tracker with the given type and name."""
  issue_tracker_project_name = _get_issue_tracker_project_name(testcase)
  if not issue_tracker_project_name or issue_tracker_project_name == 'disabled':
    return None

  return get_issue_tracker(issue_tracker_project_name)


def get_issue_tracker_policy_for_testcase(testcase):
  """Get the issue tracker with the given type and name."""
  issue_tracker_project_name = _get_issue_tracker_project_name(testcase)
  if not issue_tracker_project_name or issue_tracker_project_name == 'disabled':
    return None

  return issue_tracker_policy.get(issue_tracker_project_name)


def get_issue_for_testcase(testcase):
  """Return issue object associated with testcase."""
  if not testcase.bug_information:
    # Do not check |testcase.group_bug_information| as we look for an issue
    # associated with the testcase directly, not through a group of testcases.
    return None

  issue_tracker = get_issue_tracker_for_testcase(testcase)
  if not issue_tracker:
    return None

  issue_id = testcase.bug_information
  return issue_tracker.get_original_issue(issue_id)


def get_search_keywords(testcase):
  """Get search keywords for a testcase."""
  crash_state_lines = testcase.crash_state.splitlines()
  # Use top 2 frames for searching.
  return crash_state_lines[:2]


def get_similar_issues(issue_tracker, testcase, only_open=True):
  """Get issue objects that seem to be related to a particular test case."""
  # Get list of issues using the search query.
  keywords = get_search_keywords(testcase)

  issues = issue_tracker.find_issues(keywords=keywords, only_open=only_open)
  if issues:
    issues = list(issues)
  else:
    issues = []

  issue_ids = [issue.id for issue in issues]

  # Add issues from similar testcases sharing the same group id.
  if testcase.group_id:
    group_query = data_types.Testcase.query(
        data_types.Testcase.group_id == testcase.group_id)
    similar_testcases = ndb_utils.get_all_from_query(group_query)
    for similar_testcase in similar_testcases:
      if not similar_testcase.bug_information:
        continue

      # Exclude issues already added above from search terms.
      issue_id = int(similar_testcase.bug_information)
      if issue_id in issue_ids:
        continue

      # Get issue object using ID.
      issue = issue_tracker.get_issue(issue_id)
      if not issue:
        continue

      # If our search criteria allows open bugs only, then check issue and
      # testcase status so as to exclude closed ones.
      if (only_open and (not issue.is_open or not testcase.open)):
        continue

      issues.append(issue)
      issue_ids.append(issue_id)

  return issues


def get_similar_issues_url(issue_tracker, testcase, only_open=True):
  """Get similar issues web URL."""
  keywords = get_search_keywords(testcase)
  return issue_tracker.find_issues_url(keywords=keywords, only_open=only_open)


def get_issue_url(testcase):
  """Return issue url for a testcase. This is used when rendering a testcase,
  details page, therefore it accounts for |group_bug_information| as well."""
  issue_tracker = get_issue_tracker_for_testcase(testcase)
  if not issue_tracker:
    return None

  issue_id = (
      testcase.bug_information
      if testcase.bug_information else testcase.group_bug_information)
  if not issue_id:
    return None

  # Use str(issue_id) as |group_bug_information| might be an integer.
  return issue_tracker.issue_url(str(issue_id))


def was_label_added(issue, label):
  """Check if a label was ever added to an issue."""
  if not label:
    return False

  # Optimization that does not require pulling in issue's actions.
  if any(label.lower() == l.lower() for l in issue.labels):
    return True

  for action in issue.actions:
    for added in action.labels.added:
      if label.lower() == added.lower():
        return True

  return False



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test to see if test cases are fixed."""

import os
import time

import six

from clusterfuzz._internal.base import bisection
from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.bot.tasks import task_creation
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.chrome import crash_uploader
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.fuzzing import corpus_manager
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment


def _write_to_bigquery(testcase, progression_range_start,
                       progression_range_end):
  """Write the fixed range to BigQuery."""
  big_query.write_range(
      table_id='fixeds',
      testcase=testcase,
      range_name='fixed',
      start=progression_range_start,
      end=progression_range_end)


def _log_output(revision, crash_result):
  """Log process output."""
  logs.log(
      'Testing %s.' % revision,
      revision=revision,
      output=crash_result.get_stacktrace(symbolized=True))


def _check_fixed_for_custom_binary(testcase, job_type, testcase_file_path):
  """Simplified fixed check for test cases using custom binaries."""
  revision = environment.get_value('APP_REVISION')

  # Update comments to reflect bot information and clean up old comments.
  testcase_id = testcase.key.id()
  testcase = data_handler.get_testcase_by_id(testcase_id)
  data_handler.update_testcase_comment(testcase, data_types.TaskState.STARTED)

  build_manager.setup_build()
  if not build_manager.check_app_path():
    testcase = data_handler.get_testcase_by_id(testcase_id)
    data_handler.update_testcase_comment(
        testcase, data_types.TaskState.ERROR,
        'Build setup failed for custom binary')
    build_fail_wait = environment.get_value('FAIL_WAIT')
    tasks.add_task(
        'progression', testcase_id, job_type, wait_time=build_fail_wait)
    return

  test_timeout = environment.get_value('TEST_TIMEOUT', 10)
  result = testcase_manager.test_for_crash_with_retries(
      testcase, testcase_file_path, test_timeout, http_flag=testcase.http_flag)
  _log_output(revision, result)

  # Re-fetch to finalize testcase updates in branches below.
  testcase = data_handler.get_testcase_by_id(testcase.key.id())

  # If this still crashes on the most recent build, it's not fixed. The task
  # will be rescheduled by a cron job and re-attempted eventually.
  if result.is_crash():
    app_path = environment.get_value('APP_PATH')
    command = testcase_manager.get_command_line_for_application(
        testcase_file_path, app_path=app_path, needs_http=testcase.http_flag)
    symbolized_crash_stacktrace = result.get_stacktrace(symbolized=True)
    unsymbolized_crash_stacktrace = result.get_stacktrace(symbolized=False)
    stacktrace = utils.get_crash_stacktrace_output(
        command, symbolized_crash_stacktrace, unsymbolized_crash_stacktrace)
    testcase.last_tested_crash_stacktrace = data_handler.filter_stacktrace(
        stacktrace)
    data_handler.update_progression_completion_metadata(
        testcase,
        revision,
        is_crash=True,
        message='still crashes on latest custom build')
    return

  # Retry once on another bot to confirm our results and in case this bot is in
  # a bad state which we didn't catch through our usual means.
  if data_handler.is_first_retry_for_task(testcase, reset_after_retry=True):
    tasks.add_task('progression', testcase_id, job_type)
    data_handler.update_progression_completion_metadata(testcase, revision)
    return

  # The bug is fixed.
  testcase.fixed = 'Yes'
  testcase.open = False
  data_handler.update_progression_completion_metadata(
      testcase, revision, message='fixed on latest custom build')


def _update_issue_metadata(testcase):
  """Update issue metadata."""
  if testcase.uploader_email:
    # Trust the uploader specified metadata.
    return

  metadata = engine_common.get_all_issue_metadata_for_testcase(testcase)
  if not metadata:
    return

  for key, value in six.iteritems(metadata):
    old_value = testcase.get_metadata(key)
    if old_value != value:
      logs.log('Updating issue metadata for {} from {} to {}.'.format(
          key, old_value, value))
      testcase.set_metadata(key, value)


def _testcase_reproduces_in_revision(testcase,
                                     testcase_file_path,
                                     job_type,
                                     revision,
                                     update_metadata=False):
  """Test to see if a test case reproduces in the specified revision."""
  build_manager.setup_build(revision)
  if not build_manager.check_app_path():
    raise errors.BuildSetupError(revision, job_type)

  if testcase_manager.check_for_bad_build(job_type, revision):
    log_message = 'Bad build at r%d. Skipping' % revision
    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    data_handler.update_testcase_comment(testcase, data_types.TaskState.WIP,
                                         log_message)
    raise errors.BadBuildError(revision, job_type)

  test_timeout = environment.get_value('TEST_TIMEOUT', 10)
  result = testcase_manager.test_for_crash_with_retries(
      testcase, testcase_file_path, test_timeout, http_flag=testcase.http_flag)
  _log_output(revision, result)

  if update_metadata:
    _update_issue_metadata(testcase)

  return result


def _save_current_fixed_range_indices(testcase_id, fixed_range_start,
                                      fixed_range_end):
  """Save current fixed range indices in case we die in middle of task."""
  testcase = data_handler.get_testcase_by_id(testcase_id)
  testcase.set_metadata(
      'last_progression_min', fixed_range_start, update_testcase=False)
  testcase.set_metadata(
      'last_progression_max', fixed_range_end, update_testcase=False)
  testcase.put()


def _save_fixed_range(testcase_id, min_revision, max_revision,
                      testcase_file_path):
  """Update a test case and other metadata with a fixed range."""
  testcase = data_handler.get_testcase_by_id(testcase_id)
  testcase.fixed = '%d:%d' % (min_revision, max_revision)
  testcase.open = False

  data_handler.update_progression_completion_metadata(
      testcase, max_revision, message='fixed in range r%s' % testcase.fixed)
  _write_to_bigquery(testcase, min_revision, max_revision)

  _store_testcase_for_regression_testing(testcase, testcase_file_path)


def _store_testcase_for_regression_testing(testcase, testcase_file_path):
  """Stores reproduction testcase for future regression testing in corpus
  pruning task."""
  if testcase.open:
    # Store testcase only after the crash is fixed.
    return

  if not testcase.bug_information:
    # Only store crashes with bugs associated with them.
    return

  fuzz_target = data_handler.get_fuzz_target(testcase.overridden_fuzzer_name)
  if not fuzz_target:
    # No work to do, only applicable for engine fuzzers.
    return

  corpus = corpus_manager.FuzzTargetCorpus(fuzz_target.engine,
                                           fuzz_target.project_qualified_name())
  regression_testcase_url = os.path.join(
      corpus.get_regressions_corpus_gcs_url(),
      utils.file_hash(testcase_file_path))

  if storage.copy_file_to(testcase_file_path, regression_testcase_url):
    logs.log('Successfully stored testcase for regression testing: ' +
             regression_testcase_url)
  else:
    logs.log_error('Failed to store testcase for regression testing: ' +
                   regression_testcase_url)


def find_fixed_range(testcase_id, job_type):
  """Attempt to find the revision range where a testcase was fixed."""
  deadline = tasks.get_task_completion_deadline()
  testcase = data_handler.get_testcase_by_id(testcase_id)
  if not testcase:
    return

  if testcase.fixed:
    logs.log_error('Fixed range is already set as %s, skip.' % testcase.fixed)
    return

  # Setup testcase and its dependencies.
  file_list, _, testcase_file_path = setup.setup_testcase(testcase, job_type)
  if not file_list:
    return

  # Set a flag to indicate we are running progression task. This shows pending
  # status on testcase report page and avoid conflicting testcase updates by
  # triage cron.
  testcase.set_metadata('progression_pending', True)

  # Custom binaries are handled as special cases.
  if build_manager.is_custom_binary():
    _check_fixed_for_custom_binary(testcase, job_type, testcase_file_path)
    return

  build_bucket_path = build_manager.get_primary_bucket_path()
  revision_list = build_manager.get_revisions_list(
      build_bucket_path, testcase=testcase)
  if not revision_list:
    data_handler.close_testcase_with_error(testcase_id,
                                           'Failed to fetch revision list')
    return

  # Use min, max_index to mark the start and end of revision list that is used
  # for bisecting the progression range. Set start to the revision where noticed
  # the crash. Set end to the trunk revision. Also, use min, max from past run
  # if it timed out.
  min_revision = testcase.get_metadata('last_progression_min')
  max_revision = testcase.get_metadata('last_progression_max')

  if min_revision or max_revision:
    # Clear these to avoid using them in next run. If this run fails, then we
    # should try next run without them to see it succeeds. If this run succeeds,
    # we should still clear them to avoid capping max revision in next run.
    testcase = data_handler.get_testcase_by_id(testcase_id)
    testcase.delete_metadata('last_progression_min', update_testcase=False)
    testcase.delete_metadata('last_progression_max', update_testcase=False)
    testcase.put()

  last_tested_revision = testcase.get_metadata('last_tested_crash_revision')
  known_crash_revision = last_tested_revision or testcase.crash_revision
  if not min_revision:
    min_revision = known_crash_revision
  if not max_revision:
    max_revision = revisions.get_last_revision_in_list(revision_list)

  min_index = revisions.find_min_revision_index(revision_list, min_revision)
  if min_index is None:
    raise errors.BuildNotFoundError(min_revision, job_type)
  max_index = revisions.find_max_revision_index(revision_list, max_revision)
  if max_index is None:
    raise errors.BuildNotFoundError(max_revision, job_type)

  testcase = data_handler.get_testcase_by_id(testcase_id)
  data_handler.update_testcase_comment(testcase, data_types.TaskState.STARTED,
                                       'r%d' % max_revision)

  # Check to see if this testcase is still crashing now. If it is, then just
  # bail out.
  result = _testcase_reproduces_in_revision(
      testcase,
      testcase_file_path,
      job_type,
      max_revision,
      update_metadata=True)
  if result.is_crash():
    logs.log('Found crash with same signature on latest revision r%d.' %
             max_revision)
    app_path = environment.get_value('APP_PATH')
    command = testcase_manager.get_command_line_for_application(
        testcase_file_path, app_path=app_path, needs_http=testcase.http_flag)
    symbolized_crash_stacktrace = result.get_stacktrace(symbolized=True)
    unsymbolized_crash_stacktrace = result.get_stacktrace(symbolized=False)
    stacktrace = utils.get_crash_stacktrace_output(
        command, symbolized_crash_stacktrace, unsymbolized_crash_stacktrace)
    testcase = data_handler.get_testcase_by_id(testcase_id)
    testcase.last_tested_crash_stacktrace = data_handler.filter_stacktrace(
        stacktrace)
    data_handler.update_progression_completion_metadata(
        testcase,
        max_revision,
        is_crash=True,
        message='still crashes on latest revision r%s' % max_revision)

    # Since we've verified that the test case is still crashing, clear out any
    # metadata indicating potential flake from previous runs.
    task_creation.mark_unreproducible_if_flaky(testcase, False)

    # For chromium project, save latest crash information for later upload
    # to chromecrash/.
    state = result.get_symbolized_data()
    crash_uploader.save_crash_info_if_needed(testcase_id, max_revision,
                                             job_type, state.crash_type,
                                             state.crash_address, state.frames)
    return

  # Don't burden NFS server with caching these random builds.
  environment.set_value('CACHE_STORE', False)

  # Verify that we do crash in the min revision. This is assumed to be true
  # while we are doing the bisect.
  result = _testcase_reproduces_in_revision(testcase, testcase_file_path,
                                            job_type, min_revision)
  if result and not result.is_crash():
    testcase = data_handler.get_testcase_by_id(testcase_id)

    # Retry once on another bot to confirm our result.
    if data_handler.is_first_retry_for_task(testcase, reset_after_retry=True):
      tasks.add_task('progression', testcase_id, job_type)
      error_message = (
          'Known crash revision %d did not crash, will retry on another bot to '
          'confirm result' % known_crash_revision)
      data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                           error_message)
      data_handler.update_progression_completion_metadata(
          testcase, max_revision)
      return

    data_handler.clear_progression_pending(testcase)
    error_message = (
        'Known crash revision %d did not crash' % known_crash_revision)
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         error_message)
    task_creation.mark_unreproducible_if_flaky(testcase, True)
    return

  # Start a binary search to find last non-crashing revision. At this point, we
  # know that we do crash in the min_revision, and do not crash in max_revision.
  while time.time() < deadline:
    min_revision = revision_list[min_index]
    max_revision = revision_list[max_index]

    # If the min and max revisions are one apart this is as much as we can
    # narrow the range.
    if max_index - min_index == 1:
      _save_fixed_range(testcase_id, min_revision, max_revision,
                        testcase_file_path)
      return

    # Occasionally, we get into this bad state. It seems to be related to test
    # cases with flaky stacks, but the exact cause is unknown.
    if max_index - min_index < 1:
      testcase = data_handler.get_testcase_by_id(testcase_id)
      testcase.fixed = 'NA'
      testcase.open = False
      message = ('Fixed testing errored out (min and max revisions '
                 'are both %d)' % min_revision)
      data_handler.update_progression_completion_metadata(
          testcase, max_revision, message=message)

      # Let the bisection service know about the NA status.
      bisection.request_bisection(testcase)
      return

    # Test the middle revision of our range.
    middle_index = (min_index + max_index) // 2
    middle_revision = revision_list[middle_index]

    testcase = data_handler.get_testcase_by_id(testcase_id)
    log_message = 'Testing r%d (current range %d:%d)' % (
        middle_revision, min_revision, max_revision)
    data_handler.update_testcase_comment(testcase, data_types.TaskState.WIP,
                                         log_message)

    try:
      result = _testcase_reproduces_in_revision(testcase, testcase_file_path,
                                                job_type, middle_revision)
    except errors.BadBuildError:
      # Skip this revision.
      del revision_list[middle_index]
      max_index -= 1
      continue

    if result.is_crash():
      min_index = middle_index
    else:
      max_index = middle_index

    _save_current_fixed_range_indices(testcase_id, revision_list[min_index],
                                      revision_list[max_index])

  # If we've broken out of the loop, we've exceeded the deadline. Recreate the
  # task to pick up where we left off.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  error_message = ('Timed out, current range r%d:r%d' %
                   (revision_list[min_index], revision_list[max_index]))
  data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                       error_message)
  tasks.add_task('progression', testcase_id, job_type)


def execute_task(testcase_id, job_type):
  """Execute progression task."""
  try:
    find_fixed_range(testcase_id, job_type)
  except errors.BuildSetupError as error:
    # If we failed to setup a build, it is likely a bot error. We can retry
    # the task in this case.
    testcase = data_handler.get_testcase_by_id(testcase_id)
    error_message = 'Build setup failed r%d' % error.revision
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         error_message)
    build_fail_wait = environment.get_value('FAIL_WAIT')
    tasks.add_task(
        'progression', testcase_id, job_type, wait_time=build_fail_wait)
  except errors.BadBuildError:
    # Though bad builds when narrowing the range are recoverable, certain builds
    # being marked as bad may be unrecoverable. Recoverable ones should not
    # reach this point.
    testcase = data_handler.get_testcase_by_id(testcase_id)
    error_message = 'Unable to recover from bad build'
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         error_message)

  # If there is a fine grained bisection service available, request it. Both
  # regression and fixed ranges are requested once. Regression is also requested
  # here as the bisection service may require details that are not yet available
  # (e.g. issue ID) at the time regress_task completes.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  bisection.request_bisection(testcase)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Process handling utilities."""

from distutils import spawn
import os
import subprocess
import sys
import tempfile
import threading
import time

try:
  import psutil
except ImportError:
  psutil = None

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment


def _end_process(terminate_function, process_result):
  """Ends a running process.

  Ignores exceptions.

  Args:
    process: A subprocess.Popen object.
    terminate_function: The function to terminate the process.
    process_result: A ProcessResult object where timeout information will be
        written to.
  """
  try:
    terminate_function()
  except OSError:
    logs.log('Process already killed.')

  process_result.timed_out = True


def wait_process(process,
                 timeout,
                 input_data=None,
                 terminate_before_kill=False,
                 terminate_wait_time=None):
  """Waits until either the process exits or times out.

  Args:
    process: A subprocess.Popen object.
    timeout: Maximum number of seconds to wait for before sending a signal.
    input_data: Input to be sent to the process.
    terminate_before_kill: A bool indicating that SIGTERM should be sent to
        the process first before SIGKILL (to let the SIGTERM handler run).
    terminate_wait_time: Maximum number of seconds to wait for the SIGTERM
        handler.

  Returns:
    A ProcessResult.
  """
  result = ProcessResult()
  is_windows = environment.platform() == 'WINDOWS'

  # On Windows, terminate() just calls Win32 API function TerminateProcess()
  # which is equivalent to process kill. So, skip terminate_before_kill.
  if terminate_before_kill and not is_windows:
    first_timeout_function = process.terminate

    # Use a second timer to send the process kill.
    second_timer = threading.Timer(timeout + terminate_wait_time, _end_process,
                                   [process.kill, result])
  else:
    first_timeout_function = process.kill
    second_timer = None

  first_timer = threading.Timer(timeout, _end_process,
                                [first_timeout_function, result])

  output = None
  start_time = time.time()

  try:
    first_timer.start()
    if second_timer:
      second_timer.start()

    output = process.communicate(input_data)[0]
  finally:
    first_timer.cancel()

    if second_timer:
      second_timer.cancel()

  result.return_code = process.poll()
  result.output = output
  result.time_executed = time.time() - start_time
  return result


def kill_process_tree(root_pid):
  """Kill process tree."""
  try:
    parent = psutil.Process(root_pid)
    children = parent.children(recursive=True)
  except (psutil.AccessDenied, psutil.NoSuchProcess, OSError):
    logs.log_warn('Failed to find or access process.')
    return

  for child in children:
    try:
      child.kill()
    except (psutil.AccessDenied, psutil.NoSuchProcess, OSError):
      logs.log_warn('Failed to kill process child.')

  try:
    parent.kill()
  except (psutil.AccessDenied, psutil.NoSuchProcess, OSError):
    logs.log_warn('Failed to kill process.')


class ChildProcess(object):
  """A class representing a process that's running."""

  def __init__(self,
               popen,
               command,
               max_stdout_len=None,
               stdout_file=None,
               interactive=False):
    self._popen = popen
    self._command = command
    self._max_stdout_len = max_stdout_len
    self._stdout_file = stdout_file
    self._interactive = interactive

  @property
  def command(self):
    return self._command

  @property
  def popen(self):
    return self._popen

  def communicate(self, input=None):  # pylint: disable=redefined-builtin
    """subprocess.Popen.communicate."""
    stdout = b''
    stderr = b''

    if self._interactive:
      if input:
        self._popen.stdin.write(input)

      while True:
        line = self._popen.stdout.readline()
        if not line:
          break

        if self._stdout_file:
          self._stdout_file.write(line)
        else:
          stdout += line

        sys.stdout.write(line.decode())

      self._popen.wait()
    else:
      stdout, stderr = self._popen.communicate(input)

    if not self._max_stdout_len:
      return stdout, stderr

    with self._stdout_file:
      return utils.read_from_handle_truncated(self._stdout_file,
                                              self._max_stdout_len), stderr

  def poll(self):
    """subprocess.Popen.poll."""
    return self._popen.poll()

  def kill(self):
    """Kills running process and all of its associated children."""
    kill_process_tree(self._popen.pid)

  def terminate(self):
    """subprocess.Popen.terminate."""
    try:
      self._popen.terminate()
    except OSError:
      logs.log_warn('Failed to terminate process.')


class ProcessResult(object):
  """Object representing result of a process execution.

  Returned by ProcessRunner.run_and_wait().

  Attributes:
    command: A list of arguments representing the command line that was run.
    return_code: Exit code of the process.
    output: Process output.
    time_executed: Number of seconds process ran for.
    timed_out: Whether or not the process timed out.
  """

  def __init__(self,
               command=None,
               return_code=None,
               output=None,
               time_executed=None,
               timed_out=False):
    """Inits the ProcessResult."""
    self.command = command
    self.return_code = return_code
    self.output = output
    self.time_executed = time_executed
    self.timed_out = timed_out


class ProcessRunner(object):
  """Generic process runner class.

  Attributes:
    executable_path: Path to the executable to be run.
    default_args: An optional sequence of arguments that are always passed to
        the executable when run.
  """

  def __init__(self, executable_path, default_args=None):
    """Inits ProcessRunner."""
    self._executable_path = executable_path
    self._default_args = []

    if default_args:
      self.default_args.extend(default_args)

  @property
  def executable_path(self):
    return self._executable_path

  @property
  def default_args(self):
    return self._default_args

  def get_command(self, additional_args=None):
    """Returns the command line for running the executable.

    Args:
      additional_args: A sequence of additional arguments to be passed to the
          executable.

    Returns:
      A list containing the command arguments to be passed to subprocess.Popen.
    """
    command = [self._executable_path]
    command.extend(self._default_args)

    if additional_args:
      command.extend(additional_args)

    return command

  def run(self,
          additional_args=None,
          max_stdout_len=None,
          extra_env=None,
          stdin=subprocess.PIPE,
          stdout=subprocess.PIPE,
          stderr=subprocess.STDOUT,
          **popen_args):
    """Runs the executable.

    Does not block the caller.

    Args:
      additional_args: A sequence of additional arguments to be passed to the
          executable.
      max_stdout_len: Optional. Maximum number of bytes to collect in stdout.
      extra_env: Optional. A dictionary containing environment variables and
        their values. These will be set in the environment of the new process.
      stdin: Optional. Passed to subprocess.Popen, defaults to subprocess.PIPE,
      stdout: Optional. Passed to subprocess.Popen, defaults to subprocess.PIPE
      stderr: Optional. Passed to subprocess.Popen, defaults to
          subprocess.STDOUT
      **popen_args: Additional arguments that are passed to subprocess.Popen.

    Returns:
      A subprocess.Popen object for the process.
    """
    # TODO: Rename popen_args to popen_kwargs.
    command = self.get_command(additional_args)

    stdout_file = None
    if stdout == subprocess.PIPE and max_stdout_len:
      stdout_file = tempfile.TemporaryFile()
      stdout = stdout_file

    interactive = environment.get_value('CF_INTERACTIVE')
    if interactive:
      popen_args['bufsize'] = 0
      if stdout != subprocess.PIPE:
        # If the provided stdout is a file object, (i.e. not subprocess.PIPE),
        # we need to pipe writes through to there to ensure consistent
        # behaviour.
        stdout_file = stdout

      stdout = subprocess.PIPE

    env = popen_args.pop('env', os.environ.copy())
    if extra_env is not None:
      env.update(extra_env)

    return ChildProcess(
        subprocess.Popen(
            command,
            env=env,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            **popen_args),
        command,
        max_stdout_len=max_stdout_len,
        stdout_file=stdout_file,
        interactive=interactive)

  # Note: changes to this function may require changes to
  # untrusted_runner.proto.
  def run_and_wait(self,
                   additional_args=None,
                   timeout=None,
                   terminate_before_kill=False,
                   terminate_wait_time=None,
                   input_data=None,
                   max_stdout_len=None,
                   extra_env=None,
                   stdin=subprocess.PIPE,
                   stdout=subprocess.PIPE,
                   stderr=subprocess.STDOUT,
                   **popen_args):
    """Runs the executable.

    Blocks the caller until the process exits.

    Args:
      additional_args: A sequence of additional arguments to be passed to the
          executable.
      timeout: Maximum number of seconds to run the process for.
      terminate_before_kill: A bool indicating that SIGTERM should be sent to
          the process first before SIGKILL (to let the SIGTERM handler run).
      terminate_wait_time: Maximum number of seconds to wait for the SIGTERM
          handler.
      input_data: Optional. A string to be passed as input to the process.
      max_stdout_len: Optional. Maximum number of bytes to collect in stdout.
      extra_env: Optional. A dictionary containing environment variables and
           their values. These will be added to the environment of the new
           process.
      stdout: Optional. Passed to subprocess.Popen, defaults to subprocess.PIPE
      stderr: Optional. Passed to subprocess.Popen, defaults to
          subprocess.STDOUT
      **popen_args: Additional arguments that are passed to subprocess.Popen.

    Returns:
      A tuple of (return code, output, time process ran for, or None on timeout)
    """
    process = self.run(
        additional_args,
        max_stdout_len=max_stdout_len,
        extra_env=extra_env,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
        **popen_args)

    start_time = time.time()

    if not timeout:
      output = process.communicate(input_data)[0]
      return ProcessResult(process.command, process.poll(), output,
                           time.time() - start_time, False)

    result = wait_process(
        process,
        timeout=timeout,
        input_data=input_data,
        terminate_before_kill=terminate_before_kill,
        terminate_wait_time=terminate_wait_time)
    result.command = process.command

    return result


class UnicodeProcessRunnerMixin(object):
  """Mixin for process runner subclasses to output unicode output."""

  def run_and_wait(self, *args, **kwargs):  # pylint: disable=arguments-differ
    """Overridden run_and_wait which always decodes the output."""
    result = ProcessRunner.run_and_wait(self, *args, **kwargs)
    if result.output is not None:
      result.output = utils.decode_to_unicode(result.output)

    return result


class UnicodeProcessRunner(UnicodeProcessRunnerMixin, ProcessRunner):
  """ProcessRunner which always returns unicode output."""


class UnshareProcessRunnerMixin(object):
  """ProcessRunner mixin which unshares."""

  def get_command(self, additional_args=None):
    """Overridden get_command."""
    if environment.platform() != 'LINUX':
      raise RuntimeError('UnshareProcessRunner only supported on Linux')

    unshare_path = spawn.find_executable('unshare')
    if not unshare_path:
      raise RuntimeError('unshare not found')

    command = [unshare_path]
    command.extend([
        '-n',  # Enter network namespace.
    ])

    command.append(self._executable_path)
    command.extend(self._default_args)

    if additional_args:
      command.extend(additional_args)

    return command


class UnshareProcessRunner(UnshareProcessRunnerMixin, ProcessRunner):
  """ProcessRunner which unshares."""



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fake objects used in reproduce tool tests."""

from local.butler import reproduce
from local.butler.reproduce_tool import http_utils


class FakeResponse(object):
  """Fake response object."""

  def __init__(self, status, include_auth_header=False, filename=None):
    self.status = status
    self.include_auth_header = include_auth_header
    self.filename = filename

  def __contains__(self, item):
    if item == http_utils.AUTHORIZATION_HEADER and self.include_auth_header:
      return True
    if item == reproduce.FILENAME_RESPONSE_HEADER and self.filename:
      return True

    return False

  def __getitem__(self, item):
    if item == http_utils.AUTHORIZATION_HEADER and self.include_auth_header:
      return 'fake auth token'
    if item == reproduce.FILENAME_RESPONSE_HEADER and self.filename:
      return self.filename

    return None


class FakeHttp(object):
  """Fake HTTP object. Stores information on the last request."""

  def __init__(self, replies):
    self.replies = replies
    self.last_url = None
    self.last_method = None
    self.last_headers = None
    self.last_body = None

  def request(self, url, method, headers, body):
    """Return a predetermined response and content."""
    self.last_url = url
    self.last_method = method
    self.last_headers = headers
    self.last_body = body

    return self.replies.pop(0)


class FakeConfig(object):
  """Fake configuration object."""

  def get(self, key):
    """Fake get."""
    if key == 'oauth_url':
      return 'https://oauth-url/'
    if key == 'testcase_download_url':
      return 'https://clusterfuzz/testcase-detail/download-testcase'

    raise Exception('Unexpected config key access: ' + key)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for managing Fuchsia devices."""

import glob
import os
import re
import shutil
import subprocess

import six

from .host import Host


class Device(object):
  """Represents a Fuchsia device attached to a host.

    This class abstracts the details of remotely running commands and
    transferring data to and from the device.

    Attributes:
      host: A Host object represent the local platform attached to this target
        device.
    """

  @classmethod
  def from_args(cls, host, args):
    """Constructs a Device from command line arguments."""
    netaddr_cmd = ['netaddr', '--fuchsia', '--nowait']
    default_device = '{}.device'.format(host.build_dir)
    if args.device:
      netaddr_cmd.append(args.device)
    elif os.path.exists(default_device):
      with open(default_device) as f:
        netaddr_cmd.append(f.read().strip())
    try:
      netaddr = host.zircon_tool(netaddr_cmd)
    except subprocess.CalledProcessError:
      raise RuntimeError('Unable to find device; try `fx set-device`.')
    device = cls(host, netaddr)
    if not host.build_dir:
      raise Host.ConfigError('Unable to find SSH configuration.')
    device.set_ssh_config(Host.join(host.build_dir, 'ssh-keys', 'ssh_config'))
    return device

  def __init__(self, host, addr, port=22):
    self.host = host
    self._addr = addr
    self._ssh_opts = {}
    if port != 22:
      self._ssh_opts['p'] = [str(port)]

  def set_ssh_config(self, config_file):
    """Sets the SSH arguments to use a config file."""
    if not os.path.exists(config_file):
      raise Host.ConfigError('Unable to find SSH configuration.')
    self._ssh_opts['F'] = [config_file]

  def set_ssh_identity(self, identity_file):
    if not os.path.exists(identity_file):
      raise Host.ConfigError('Unable to find SSH identity.')
    self._ssh_opts['i'] = [identity_file]

  def set_ssh_option(self, option):
    """Sets SSH configuration options. Can be used multiple times."""
    if 'o' in self._ssh_opts:
      self._ssh_opts['o'].append(option)
    else:
      self._ssh_opts['o'] = [option]

  def set_ssh_verbosity(self, level):
    """Sets how much debugging SSH prints. Default is 0 (none), max is 3."""
    for i in range(1, 4):
      opt = 'v' * i
      if level == i and not opt in self._ssh_opts:
        self._ssh_opts[opt] = []
      elif level != i and opt in self._ssh_opts:
        del self._ssh_opts[opt]

  def get_ssh_cmd(self, cmd):
    """Returns the SSH executable and options."""
    result = cmd[:1]
    for opt, args in six.iteritems(self._ssh_opts):
      if result[0] == 'scp' and opt == 'p':
        opt = 'P'
      if not args:
        result.append('-' + opt)
      else:
        for arg in args:
          result.append('-' + opt)
          result.append(arg)
    return result + cmd[1:]

  def _ssh(self, cmdline, stdout=subprocess.PIPE):
    """Internal wrapper around _rexec that adds the ssh command and config.

    Don't call this directly. This method exists to be overridden in testing.

    Args:
      cmdline: List of command line arguments to execute on device
      stdout: Same as for subprocess.Popen

    Returns:
      If check was false, a subprocess.Popen object representing the running
      child process.

    Raises: A Process object.
    """
    args = self.get_ssh_cmd(['ssh', self._addr] + cmdline)
    return self.host.create_process(
        args, stdout=stdout, stderr=subprocess.STDOUT)

  def ssh(self, cmdline, quiet=True, logfile=None):
    """Runs a command to completion on the device.

    Connects to the target device and executes a shell command.  Output from
    the shell command is sent to stdout, and may optionally be saved to a file
    via the POSIX utility 'tee'.

    Args:
      cmdline: A list of command line arguments, starting with the command to
        execute.
      logfile: An optional pathname to save a copy of the command output to. The
        output will also still be sent to stdout.
    """
    if quiet:
      if logfile:
        with open(logfile, 'w') as f:
          return self._ssh(cmdline, stdout=f).call()
      return self._ssh(cmdline, stdout=Host.DEVNULL).call()

    if logfile:
      p1 = self._ssh(cmdline, stdout=subprocess.PIPE).popen()
      p2 = self.host.create_process(['tee', logfile], stdin=p1.stdout)
      return p2.check_call()
    return self._ssh(cmdline, stdout=None).call()

  def getpids(self):
    """Maps names to process IDs for running fuzzers.

    Connects to the device and checks which fuzz targets have a matching entry
    in the component list given by 'cs'.  This matches on *only* the first 32
    characters of the component manifest and package URL.  This is due to 'cs'
    being limited to returning strings of length `ZX_MAX_NAME_LEN`, as defined
    in //zircon/system/public/zircon/types.h.

    Returns:
      A dict mapping fuzz target names to process IDs. May be empty if no
      fuzzers are running.
    """
    p = self._ssh(['cs'], stdout=subprocess.PIPE).popen()
    out, _ = p.communicate()
    pids = {}
    for fuzzer in self.host.fuzzers:
      tgt = (fuzzer[1] + '.cmx')[:32]
      url = ('fuchsia-pkg://fuchsia.com/%s#meta' % fuzzer[0])[:32]
      for line in out.decode('utf-8').split('\n'):
        match = re.search(tgt + r'\[(\d+)\]: ' + url, line)
        if match:
          pids[fuzzer[1]] = int(match.group(1))
    return pids

  def ls(self, path):
    """Maps file names to sizes for the given path.

    Connects to a Fuchsia device and lists the files in a directory given by
    the provided path.  Ignore non-existent paths.

    Args:
      path: Absolute path to a directory on the device.

    Returns:
      A dict mapping file names to file sizes, or an empty dict if the path
      does not exist.
    """
    results = {}
    try:
      p = self._ssh(['ls', '-l', path], stdout=subprocess.PIPE).popen()
      out, _ = p.communicate()
      for line in out.decode('utf-8').split('\n'):
        # Line ~= '-rw-r--r-- 1 0 0 8192 Mar 18 22:02 some-name'
        parts = line.split()
        # When we're running ls over ssh, we may get a note about
        # "Warning: Permanently added [address] to the list of known hosts"
        # Don't try to treat those as file paths
        if len(parts) > 8 and 'Warning:' not in parts:
          results[' '.join(parts[8:])] = int(parts[4])
    except subprocess.CalledProcessError:
      pass
    return results

  def rm(self, pathname, recursive=False):
    """Removes a file or directory from the device."""
    args = ['rm']
    if recursive:
      args.append('-r')
    args.append(pathname)
    self.ssh(args)

  def _dump_log(self, args):
    """Retrieve a syslog from the device."""
    p = self._ssh(['log_listener', '--dump_logs', 'yes'] + args)
    return p.check_output()

  def _guess_pid(self):
    """Tries to guess the fuzzer process ID from the device syslog.

        This will assume the last line which contained one of the strings
        '{{{reset}}}', 'libFuzzer', or 'Sanitizer' is the fuzzer process, and
        try to extract its PID.

        Returns:
          The PID of the process suspected to be the fuzzer, or -1 if no
          suitable candidate was found.
        """
    out = self._dump_log(['--only', 'reset,Fuzzer,Sanitizer'])
    pid = -1
    for line in out.split(b'\n'):
      # Log lines are like '[timestamp][pid][tid][name] data'
      parts = line.split(b'][')
      if len(parts) > 2:
        pid = int(parts[1])
    return pid

  def process_logs(self, logfile, guess_pid=False, retcode=0):
    """Constructs a symbolized fuzzer log from a device.

        Merges the provided fuzzer log with the symbolized system log for the
        fuzzer process.

        Args:
          logfile: Absolute path to a fuzzer's log file.
          guess_pid: If true and the fuzzer process ID cannot be found in the
            fuzzer log, the process ID is picked from candidates in the system
            log.

        Returns:
          A list of the test artifacts (e.g. crashes) reported in the logs.
        """
    pid = -1
    pid_pattern = re.compile(br'==([0-9]+)==')
    mutation_pattern = re.compile(br'^MS: [0-9]*')
    artifacts = []
    artifact_pattern = re.compile(br'Test unit written to data/(\S*)')
    repro_pattern = re.compile(br'Running: .*')
    line_with_crash_message = None
    with open(logfile, 'rb') as log:
      with open(logfile + '.tmp', 'wb') as tmp:
        for line in log:
          # Check for a line that tells us the process ID
          match = pid_pattern.search(line)
          if match:
            line_with_crash_message = line
            pid = int(match.group(1))

          # Check for one of two things:
          # 1) a unit being dumped (e.g. a finding from a regular fuzz run)
          # 2) a nonzero return code plus a "Running: [foo]" message (which
          # indicates this is a *reproducer* run that has successfully crashed)
          repro_match = repro_pattern.search(line)
          match = mutation_pattern.search(line)
          if match or (repro_match and retcode):
            if pid <= 0 and guess_pid:
              pid = self._guess_pid()
            if pid > 0:
              raw = self._dump_log(['--pid', str(pid)])
              sym = self.host.symbolize(raw)
              tmp.write(b'\n'.join(sym))
              tmp.write(b'\n')

          # Check for an artifact being reported.
          match = artifact_pattern.search(line)
          if match:
            artifacts.append(match.group(1))

          # Echo the line
          tmp.write(line)

    # Clusterfuzz's stack analyzer expects the
    # `==[num]== ERROR: [SanitizerName]: [failure type]` line
    # to occur *before* the stacktrace, so make a new tempfile
    # where we insert that line at the top.
    # TODO(flowerhack): Change the log output in Fuchsia itself, s.t. the
    # ordering is correct the *first* time, and we won't have to do this
    # fix-up-the-logs dance!
    with open(logfile + '.tmp', 'rb') as tmp:
      with open(logfile, 'wb') as final:
        if line_with_crash_message:
          final.write(line_with_crash_message)

        shutil.copyfileobj(tmp, final)

    os.remove(logfile + '.tmp')
    return artifacts

  def _scp(self, srcs, dst):
    """Copies `src` to `dst`.

    Don't call directly; use `fetch` or `store` instead.`

    Args:
      srcs: Local or remote paths to copy from.
      dst: Local or remote path to copy to.
    """
    args = self.get_ssh_cmd(['scp'] + srcs + [dst])
    p = self.host.create_process(args)
    p.call()

  def fetch(self, data_src, host_dst):
    """Copies `data_src` on the target to `host_dst` on the host."""
    if not os.path.isdir(host_dst):
      raise ValueError(host_dst + ' is not a directory')
    self._scp(['[{}]:{}'.format(self._addr, data_src)], host_dst)

  def store(self, host_src, data_dst):
    """Copies `host_src` on the host to `data_dst` on the target."""
    self.ssh(['mkdir', '-p', data_dst])
    srcs = glob.glob(host_src)
    if not srcs:
      return
    self._scp(srcs, '[{}]:{}'.format(self._addr, data_dst))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for external_update."""

import base64
import datetime
import json
import os
import unittest

import flask
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers import external_update

DATA_DIRECTORY = os.path.join(os.path.dirname(__file__), 'external_update_data')


@test_utils.with_cloud_emulators('datastore')
class ExternalUpdatesTest(unittest.TestCase):
  """Test external updates."""

  def _read_test_data(self, name):
    """Helper function to read test data."""
    with open(os.path.join(DATA_DIRECTORY, name), 'rb') as handle:
      return handle.read()

  def setUp(self):
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/external-update', view_func=external_update.Handler.as_view(''))
    self.app = webtest.TestApp(flaskapp)

    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
        'google.oauth2.id_token.verify_oauth2_token',
    ])

    self.mock.verify_oauth2_token.return_value = {
        'email_verified': True,
        'email': 'test-clusterfuzz@appspot.gserviceaccount.com',
    }
    self.mock.utcnow.return_value = datetime.datetime(2021, 1, 1)

    data_types.Job(
        name='libfuzzer_external_job',
        external_reproduction_topic='topic',
        external_updates_subscription='subscription').put()
    data_types.Job(name='job').put()

    data_types.FuzzTarget(engine='libFuzzer', binary='fuzzer').put()

    self.testcase = data_types.Testcase(
        open=True,
        status='Processed',
        fuzzer_name='libFuzzer',
        overridden_fuzzer_name='libFuzzer_fuzzer',
        job_type='libfuzzer_external_job',
        crash_state=('blink::InputTypeView::element\n'
                     'blink::TextFieldInputType::didSetValueByUserEdit\n'
                     'blink::TextFieldInputType::subtreeHasChanged\n'),
        crash_revision=1336,
        crash_stacktrace='original',
        last_tested_crash_stacktrace='last_tested',
        crash_type='',
        security_flag=True)
    self.testcase.put()

  def _make_message(self, data, attributes):
    """Make a message."""
    return json.dumps({
        'message': {
            'data': base64.b64encode(data).decode(),
            'attributes': attributes,
        }
    })

  def test_update_still_crashing(self):
    """Test an update that is still crashing."""
    stacktrace = self._read_test_data('asan_uaf.txt')
    self.app.post(
        '/external-update',
        params=self._make_message(stacktrace, {
            'testcaseId': self.testcase.key.id(),
            'revision': '1337'
        }),
        headers={'Authorization': 'Bearer fake'},
        content_type='application/octet-stream')

    updated_testcase = self.testcase.key.get()
    self.assertTrue(updated_testcase.open)
    self.assertEqual('', updated_testcase.fixed)
    self.assertEqual(stacktrace.decode(),
                     updated_testcase.last_tested_crash_stacktrace)
    self.assertEqual(1337,
                     updated_testcase.get_metadata('last_tested_revision'))
    self.assertEqual(
        1337, updated_testcase.get_metadata('last_tested_crash_revision'))

  def test_update_changed_security(self):
    """Test an update that is still crashing, but with a different security
    flag."""
    self.testcase.security_flag = False
    self.testcase.put()

    stacktrace = self._read_test_data('asan_uaf.txt')
    self.app.post(
        '/external-update',
        params=self._make_message(stacktrace, {
            'testcaseId': self.testcase.key.id(),
            'revision': '1337'
        }),
        headers={'Authorization': 'Bearer fake'},
        content_type='application/octet-stream')

    updated_testcase = self.testcase.key.get()
    self.assertFalse(updated_testcase.open)
    self.assertEqual('1336:1337', updated_testcase.fixed)
    self.assertEqual('last_tested',
                     updated_testcase.last_tested_crash_stacktrace)
    self.assertEqual(1337,
                     updated_testcase.get_metadata('last_tested_revision'))
    self.assertIsNone(
        updated_testcase.get_metadata('last_tested_crash_revision'))

  def test_update_older_revision(self):
    """Test an update that is for an older revision."""
    self.app.post(
        '/external-update',
        params=self._make_message(b'', {
            'testcaseId': self.testcase.key.id(),
            'revision': '1335'
        }),
        headers={'Authorization': 'Bearer fake'},
        content_type='application/octet-stream')

    updated_testcase = self.testcase.key.get()
    self.assertTrue(updated_testcase.open)
    self.assertEqual('', updated_testcase.fixed)
    self.assertEqual('last_tested',
                     updated_testcase.last_tested_crash_stacktrace)
    self.assertIsNone(updated_testcase.get_metadata('last_tested_revision'))
    self.assertIsNone(
        updated_testcase.get_metadata('last_tested_crash_revision'))

  def test_update_fixed(self):
    """Test an update that is no longer crashing."""
    self.app.post(
        '/external-update',
        params=self._make_message(b'', {
            'testcaseId': self.testcase.key.id(),
            'revision': '1337'
        }),
        headers={'Authorization': 'Bearer fake'},
        content_type='application/octet-stream')

    updated_testcase = self.testcase.key.get()
    self.assertFalse(updated_testcase.open)
    self.assertEqual('1336:1337', updated_testcase.fixed)
    self.assertEqual('last_tested',
                     updated_testcase.last_tested_crash_stacktrace)
    self.assertEqual(1337,
                     updated_testcase.get_metadata('last_tested_revision'))
    self.assertIsNone(
        updated_testcase.get_metadata('last_tested_crash_revision'))

  def test_update_error(self):
    """Test an update that has errored out."""
    self.app.post(
        '/external-update',
        params=self._make_message(
            b'', {
                'testcaseId': self.testcase.key.id(),
                'revision': '1337',
                'error': 'error'
            }),
        headers={'Authorization': 'Bearer fake'},
        content_type='application/octet-stream')

    updated_testcase = self.testcase.key.get()
    self.assertFalse(updated_testcase.open)
    self.assertEqual('NA', updated_testcase.fixed)
    self.assertEqual('last_tested',
                     updated_testcase.last_tested_crash_stacktrace)
    self.assertEqual(1337,
                     updated_testcase.get_metadata('last_tested_revision'))
    self.assertIsNone(
        updated_testcase.get_metadata('last_tested_crash_revision'))

  def test_update_not_external(self):
    """Test trying to update a testcase that isn't external."""
    self.testcase.job_type = 'job'
    self.testcase.put()
    stacktrace = self._read_test_data('asan_uaf.txt')
    resp = self.app.post(
        '/external-update',
        params=self._make_message(stacktrace, {
            'testcaseId': self.testcase.key.id(),
            'revision': '1337'
        }),
        headers={'Authorization': 'Bearer fake'},
        content_type='application/octet-stream',
        expect_errors=True)
    self.assertEqual(400, resp.status_int)

  def test_update_records_fuzz_target(self):
    """Test that updating records the fuzz target."""
    stacktrace = self._read_test_data('asan_uaf.txt')
    self.app.post(
        '/external-update',
        params=self._make_message(stacktrace, {
            'testcaseId': self.testcase.key.id(),
            'revision': '1337'
        }),
        headers={'Authorization': 'Bearer fake'},
        content_type='application/octet-stream')

    fuzz_target_job = data_types.FuzzTargetJob.get_by_id(
        'libFuzzer_fuzzer/libfuzzer_external_job')
    self.assertIsNotNone(fuzz_target_job)
    self.assertEqual(datetime.datetime(2021, 1, 1), fuzz_target_job.last_run)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test helpers for untrusted_runner."""

import os
import shutil
import subprocess
import tempfile
import unittest

from clusterfuzz._internal.bot.tasks import commands
from clusterfuzz._internal.bot.untrusted_runner import file_host
from clusterfuzz._internal.bot.untrusted_runner import host
from clusterfuzz._internal.bot.untrusted_runner import untrusted
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils

TEST_LIBS_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_LIBS_DATA_DIR = os.path.join(TEST_LIBS_DIR, 'data')


def untrusted_process():
  """Start an untrusted process."""
  os.environ['BOT_NAME'] = 'localhost'
  untrusted.start_server()


def _test_data_dir():
  """Return path to directory for bot and server data."""
  root_dir = os.environ['ROOT_DIR']
  return os.path.join(root_dir, '_test_data')


def _create_test_bot():
  """Start test bot."""
  # TODO(ochang): Use Docker container instead.
  bot_path = os.path.join(_test_data_dir(), 'worker_bot')
  if os.path.exists(bot_path):
    shutil.rmtree(bot_path, ignore_errors=True)

  env = os.environ.copy()
  env['UNTRUSTED_WORKER'] = 'True'
  env['BOT_NAME'] = 'localhost'
  bot_proc = subprocess.Popen(
      ['python', 'butler.py', 'run_bot', bot_path], env=env)

  return bot_proc, os.path.join(bot_path, 'clusterfuzz')


def _create_test_root():
  """Create test ROOT_DIR for the trusted host."""
  root_path = os.path.join(_test_data_dir(), 'test_root')
  if os.path.exists(root_path):
    shutil.rmtree(root_path, ignore_errors=True)

  real_root = os.environ['ROOT_DIR']
  os.makedirs(root_path)

  # TODO(ochang): Make sure we don't copy files that aren't tracked in git.
  shutil.copytree(
      os.path.join(real_root, 'bot'), os.path.join(root_path, 'bot'))
  shutil.copytree(
      os.path.join(real_root, 'resources'), os.path.join(
          root_path, 'resources'))

  os.mkdir(os.path.join(root_path, 'src'))
  shutil.copytree(
      os.path.join(real_root, 'src', 'appengine'),
      os.path.join(root_path, 'src', 'appengine'))
  shutil.copytree(
      os.path.join(real_root, 'src', 'python'),
      os.path.join(root_path, 'src', 'python'))
  shutil.copytree(
      os.path.join(real_root, 'src', 'clusterfuzz'),
      os.path.join(root_path, 'src', 'clusterfuzz'))
  shutil.copytree(
      os.path.join(real_root, 'src', 'third_party'),
      os.path.join(root_path, 'src', 'third_party'))

  return root_path


def _which(prog):
  """Return full path to |prog| (based on $PATH)."""
  for path in os.getenv('PATH', '').split(':'):
    candidate = os.path.join(path, prog)
    if os.path.exists(candidate):
      return candidate

  return None


@unittest.skipIf(not os.getenv('UNTRUSTED_RUNNER_TESTS'),
                 'Skipping untrusted runner tests.')
@test_utils.with_cloud_emulators('datastore', 'pubsub')
class UntrustedRunnerIntegrationTest(unittest.TestCase):
  """Base class for doing integration testing of untrusted_runner."""

  @classmethod
  def setUpClass(cls):
    cls.saved_env = os.environ.copy()
    os.environ['HOST_INSTANCE_NAME'] = 'host'
    os.environ['HOST_INSTANCE_NUM'] = '0'
    os.environ['BOT_NAME'] = 'host-0'
    os.environ['LOCAL_DEVELOPMENT'] = 'True'
    os.environ['SOURCE_VERSION_OVERRIDE'] = 'VERSION'
    os.environ['CONFIG_DIR_OVERRIDE'] = os.path.abspath(
        os.path.join(os.environ['ROOT_DIR'], 'configs', 'test'))

    cert_location = os.path.join(TEST_LIBS_DATA_DIR, 'untrusted_cert.pem')
    key_location = os.path.join(TEST_LIBS_DATA_DIR, 'untrusted_key.pem')
    os.environ['UNTRUSTED_TLS_CERT_FOR_TESTING'] = cert_location
    os.environ['UNTRUSTED_TLS_KEY_FOR_TESTING'] = key_location

    cls.bot_proc, bot_root_dir = _create_test_bot()

    os.environ['TRUSTED_HOST'] = 'True'
    os.environ['WORKER_ROOT_DIR'] = bot_root_dir
    os.environ['WORKER_BOT_TMPDIR'] = os.path.join(bot_root_dir, 'bot_tmpdir')

    environment.set_default_vars()

    data_types.HostWorkerAssignment(
        host_name='host',
        instance_num=0,
        worker_name='localhost',
        project_name='project',
        id='host-0').put()

    with open(cert_location, 'rb') as f:
      cert_contents = f.read()

    with open(key_location, 'rb') as f:
      key_contents = f.read()

    data_types.WorkerTlsCert(
        project_name='project',
        cert_contents=cert_contents,
        key_contents=key_contents,
        id='project').put()

    host.init()

  @classmethod
  def tearDownClass(cls):
    if cls.bot_proc:
      try:
        cls.bot_proc.terminate()
      except OSError:
        # Could already be killed.
        pass

    os.environ.clear()
    os.environ.update(cls.saved_env)

  def setUp(self):
    self.tmp_dir = tempfile.mkdtemp()
    os.environ['BOT_TMPDIR'] = os.path.join(self.tmp_dir, 'bot_tmpdir')

    test_helpers.patch(self, [
        'clusterfuzz._internal.datastore.data_handler.'
        'get_data_bundle_bucket_name',
        'clusterfuzz._internal.system.environment.'
        'set_environment_parameters_from_file',
    ])

    test_helpers.patch_environ(self)

    # Our tests write data/logs into subdirs of ROOT_DIR. Pivot the ROOT_DIR to
    # a temporary one.
    new_root = _create_test_root()

    os.environ['ROOT_DIR'] = new_root
    self.saved_cwd = os.getcwd()
    os.chdir(new_root)

    environment.set_bot_environment()

    fuzz_inputs = os.environ['FUZZ_INPUTS']
    shell.remove_directory(fuzz_inputs, recreate=True)

    worker_fuzz_inputs = file_host.rebase_to_worker_root(fuzz_inputs)
    shell.remove_directory(worker_fuzz_inputs, recreate=True)

    environment.set_value('GSUTIL_PATH', os.path.dirname(_which('gsutil')))

    test_utils.setup_pubsub('test-clusterfuzz')
    test_utils.create_pubsub_topic(pubsub.PubSubClient(), 'test-clusterfuzz',
                                   'jobs-project-linux')

  def tearDown(self):
    shutil.rmtree(self.tmp_dir)
    os.chdir(self.saved_cwd)

  def _setup_env(self, job_type=None):
    """Set up bot environment."""
    if not job_type:
      return

    job = data_types.Job.query(data_types.Job.name == job_type).get()
    environment.set_value('JOB_NAME', job_type)
    commands.update_environment_for_job(job.environment_string)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""data_types tests."""
import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class TestcaseTest(unittest.TestCase):
  """Test Testcase."""

  def test_put(self):
    """Test put(). It should tokenize certain fields."""
    testcase = data_types.Testcase()
    testcase.crash_state = 'state'
    testcase.crash_type = 'type'
    testcase.fuzzer_name = 'fuzzer'
    testcase.overridden_fuzzer_name = 'Overfuzzer'
    testcase.job_type = 'job'
    testcase.bug_information = '333'
    testcase.group_id = 1234
    testcase.group_bug_information = 999
    testcase.impact_extended_stable_version = 'es.1'
    testcase.impact_stable_version = 's.1'
    testcase.impact_beta_version = 'b.3'
    testcase.platform_id = 'windows'
    testcase.project_name = 'chromium'
    testcase.one_time_crasher_flag = False
    testcase.is_impact_set_flag = True
    testcase.put()

    testcase = testcase.key.get()

    self.assertSetEqual(
        set(['state', 'type', 'job', 'fuzzer', 'overfuzzer', 'windows']),
        set(testcase.keywords))
    self.assertSetEqual(set(['333', '999']), set(testcase.bug_indices))
    self.assertTrue(testcase.has_bug_flag)
    self.assertSetEqual(
        set(['fuzzer', 'Overfuzzer']), set(testcase.fuzzer_name_indices))
    self.assertSetEqual(
        set(['s', 's.1']), set(testcase.impact_stable_version_indices))
    self.assertSetEqual(
        set(['b', 'b.3']), set(testcase.impact_beta_version_indices))
    self.assertSetEqual(
        set(['es', 'es.1']),
        set(testcase.impact_extended_stable_version_indices))
    self.assertSetEqual(
        set([
            'es', 'es.1', 's', 's.1', 'b', 'b.3', 'stable', 'beta',
            'extended_stable'
        ]), set(testcase.impact_version_indices))

  def test_put_head(self):
    """Tests put() when the impact is head."""
    testcase = data_types.Testcase()
    testcase.impact_extended_stable_version = ''
    testcase.impact_stable_version = ''
    testcase.impact_beta_version = ''
    testcase.project_name = 'chromium'
    testcase.one_time_crasher_flag = False
    testcase.is_impact_set_flag = True
    testcase.put()

    testcase = testcase.key.get()

    self.assertSetEqual(set(['head']), set(testcase.impact_version_indices))

  def test_chromium_extended_stable(self):
    """Test put() when only impact_extended_stable_version is set."""
    testcase = data_types.Testcase()
    testcase.impact_extended_stable_version = 'es.1'
    testcase.impact_stable_version = ''
    testcase.impact_beta_version = ''
    testcase.project_name = 'chromium'
    testcase.one_time_crasher_flag = False
    testcase.is_impact_set_flag = True
    testcase.put()

    testcase = testcase.key.get()

    self.assertSetEqual(
        set(['es', 'es.1']),
        set(testcase.impact_extended_stable_version_indices))
    self.assertSetEqual(
        set(['es', 'es.1', 'extended_stable']),
        set(testcase.impact_version_indices))

  def test_non_chromium(self):
    """Test put(). It should tokenize certain fields."""
    testcase = data_types.Testcase()
    testcase.impact_version_indices = ['head']
    testcase.impact_stable_version = '4.5.6'
    testcase.impact_beta_version = '1.2.3'
    testcase.impact_stable_version_indices = ['s']
    testcase.impact_beta_version_indices = ['b']
    testcase.impact_stable_version_likely = True
    testcase.impact_beta_version_likely = True
    testcase.is_impact_set_flag = True
    testcase.project_name = 'cobalt'
    testcase.put()

    testcase = testcase.key.get()

    self.assertEqual([], testcase.impact_stable_version_indices)
    self.assertEqual([], testcase.impact_beta_version_indices)
    self.assertEqual([], testcase.impact_version_indices)
    # We only clear the indices. The original data is kept.
    self.assertEqual('1.2.3', testcase.impact_beta_version)
    self.assertEqual('4.5.6', testcase.impact_stable_version)
    self.assertTrue(testcase.is_impact_set_flag)
    self.assertTrue(testcase.impact_stable_version_likely)
    self.assertTrue(testcase.impact_beta_version_likely)


class FuzzTargetFullyQualifiedNameTest(unittest.TestCase):
  """Test fuzz_target_fully_qualified_name."""

  def test_project_with_regular_chars(self):
    self.assertEqual(
        'libFuzzer_myproject_json_fuzzer',
        data_types.fuzz_target_fully_qualified_name('libFuzzer', 'myproject',
                                                    'json_fuzzer'))
    self.assertEqual(
        'afl_test_project_hash_fuzzer',
        data_types.fuzz_target_fully_qualified_name('afl', 'test_project',
                                                    'hash_fuzzer'))

  def test_project_with_special_chars(self):
    self.assertEqual(
        'libFuzzer_third_party-llvm_clang_fuzzer',
        data_types.fuzz_target_fully_qualified_name(
            'libFuzzer', '//third_party/llvm', 'clang_fuzzer'))
    self.assertEqual(
        'afl_third_party-aspell-aspell_5_aspell_fuzzer',
        data_types.fuzz_target_fully_qualified_name(
            'afl', 'third_party:aspell:aspell_5', 'aspell_fuzzer'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Home page handler."""

from clusterfuzz._internal.base import external_users
from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment
from handlers import base_handler
from libs import access
from libs import handler
from libs import helpers

MEMCACHE_TTL_IN_SECONDS = 30 * 60


def _sort_by_name(item):
  """Sort key function."""
  return item['name']


def _get_engine_names(job_name):
  """Return the (engine display name, engine name) for the job."""
  if job_name.startswith('afl_'):
    return 'AFL', 'afl'

  if job_name.startswith('libfuzzer_'):
    return 'libFuzzer', 'libFuzzer'

  return 'Unknown', 'Unknown'


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def get_single_fuzz_target_or_none(project, engine_name):
  """Return the name of the single child fuzzer for the project, or None."""
  fuzz_targets = data_handler.get_fuzz_targets(
      engine=engine_name, project=project)
  fuzz_target_name = None

  for fuzz_target in fuzz_targets:
    if fuzz_target_name:
      # More than 1 child fuzzer.
      return None

    fuzz_target_name = fuzz_target.fully_qualified_name()

  return fuzz_target_name


def _get_project_results_for_jobs(jobs):
  """Return projects for jobs."""
  projects = {}
  for job in sorted(jobs, key=lambda j: j.name):
    project_name = job.get_environment().get('PROJECT_NAME', job.name)
    if project_name not in projects:
      projects[project_name] = {'name': project_name, 'jobs': []}

    if utils.string_is_true(job.get_environment().get('CORPUS_PRUNE')):
      projects[project_name]['coverage_job'] = job.name

    engine_display_name, engine_name = _get_engine_names(job.name)
    projects[project_name]['jobs'].append({
        'engine_display_name':
            engine_display_name,
        'engine_name':
            engine_name,
        'sanitizer_string':
            environment.get_memory_tool_display_string(job.name),
        'name':
            job.name,
        'single_target':
            get_single_fuzz_target_or_none(project_name, engine_name),
        'has_stats':
            True
    })

  projects = list(projects.values())
  projects.sort(key=_sort_by_name)
  for project in projects:
    project['jobs'].sort(key=_sort_by_name)

  return projects


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def _get_all_project_results():
  """Return all results."""
  jobs = list(data_types.Job.query())
  return _get_project_results_for_jobs(jobs)


def _get_project_results_for_external_user(external_jobs):
  """Return results for external user."""
  jobs = list(data_types.Job.query())
  jobs = [job for job in jobs if job.name in external_jobs]
  return _get_project_results_for_jobs(jobs)


def get_results():
  """Return results."""
  is_user = access.has_access()
  user_email = helpers.get_user_email()
  external_jobs = external_users.allowed_jobs_for_user(user_email)

  is_external_user = not is_user and external_jobs
  if not is_user and not is_external_user:
    raise helpers.AccessDeniedException()

  if is_user:
    projects = _get_all_project_results()
  else:
    projects = _get_project_results_for_external_user(external_jobs)

  results = {
      'info': {
          'projects': projects,
          'is_internal_user': is_user,
      },
  }
  return results


class Handler(base_handler.Handler):
  """Home page handler."""

  @handler.get(handler.HTML)
  def get(self):
    """GET handler."""
    return self.render('oss-fuzz-home.html', get_results())


class RefreshCacheHandler(base_handler.Handler):
  """Home page handler."""

  @handler.cron()
  def get(self):
    """GET handler."""
    # pylint: disable=unexpected-keyword-arg
    _get_all_project_results(__memoize_force__=True)
